"""Service account creation, management, and registration.

Version Added:
    8.1
"""

from __future__ import annotations

import logging
from datetime import timedelta
from typing import TYPE_CHECKING
from uuid import uuid4

from django.contrib.auth.models import User
from django.db import IntegrityError
from django.utils import timezone
from django.utils.translation import gettext as _
from djblets.avatars.services import URLAvatarService
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         ATTRIBUTE_REGISTERED,
                                         DEFAULT_ERRORS,
                                         NOT_REGISTERED,
                                         UNREGISTER)
from djblets.siteconfig.models import SiteConfiguration
from typelets.symbols import UNSET

from reviewboard.accounts.errors import ServiceAccountUserError
from reviewboard.accounts.models import Profile
from reviewboard.avatars import avatar_services
from reviewboard.registries.registry import Registry
from reviewboard.webapi.models import WebAPIToken

if TYPE_CHECKING:
    from collections.abc import Mapping

    from typelets.json import JSONDictImmutable
    from typelets.symbols import Unsettable

    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


class ServiceAccount:
    """A service account used for automation and integration purposes.

    Service accounts are special entities that are allowed to perform
    operations using the Review Board API, but cannot otherwise log in,
    and do not take a seat on the license.

    They're intended for extensions and integrations that need to associate
    data with a managed user account or pass an API token to another service.

    Service accounts are created by either providing the necessary attributes
    in a subclass or when constructing an instance. In either case, the
    instance needs to be registered in :py:data:`service_account_registry`
    prior to use.

    Version Added:
        8.1
    """

    ##########################################
    # Class customization/instance variables #
    ##########################################

    #: The unique ID of the service account.
    #:
    #: This must be in :term:`slug` format.
    #:
    #: This ID will be used as the username for new accounts if
    #: :py:attr:`preferred_username` is not provided on the class or
    #: during construction.
    service_account_id: str = ''

    #: The name of this service account.
    #:
    #: This will be used for any managed users.
    name: str = 'Service Account'

    #: The e-mail address for the service account.
    #:
    #: This will be used for any managed users. If not provided, the
    #: default e-mail address for the server will be used.
    email: str = ''

    #: The preferred username for managed users.
    #:
    #: This will be used instead of :py:attr:`service_account_id` as the
    #: username used when attempting to look up or create a user account.
    preferred_username: str = ''

    #: Explicit avatars to use for managed accounts.
    #:
    #: This is a mapping of resolution indicators (e.g., ``1x``, ``2x``,
    #: ``3x``) to image URLs (absolute or relative).
    #:
    #: If not set, avatars will not be configured.
    avatar_urls: (Mapping[str, str] | None) = None

    #: The version of the profile state.
    #:
    #: Service accounts should explicitly increment this value whenever they
    #: need to ensure new state (e-mail address, name, or avatars) is
    #: saved to the managed user.
    profile_version: int = 1

    #: The expiration time for any created API tokens, in seconds.
    #:
    #: When creating a new API token for access, this will be used to
    #: set the expiration for the token.
    #:
    #: This defaults to 30 days.
    api_token_expiration_secs: (int | None) = 60 * 60 * 24 * 30

    #: The token policy document to use for any created API tokens.
    #:
    #: It's recommended that a token policy document is set that covers
    #: the minimum set of permissions needed for the service account to
    #: perform its responsibilities.
    #:
    #: To ensure older tokens with an outdated policy are not used,
    #: increment :py:attr:`api_token_version` when updating the policy.
    api_token_policy: (JSONDictImmutable | None) = None

    #: The version of the API token state.
    #:
    #: Service accounts should explicitly increment this value whenever they
    #: need to ensure a new expiration or token policy is used the next time
    #: a token is accessed.
    api_token_version: int = 1

    ######################
    # Instance variables #
    ######################

    #: Whether an explicit username is claimed during construction.
    _claim_username: bool

    #: The cached managed user for this service account.
    _user: User | None

    def __init__(
        self,
        *,
        api_token_expiration_secs: Unsettable[int] = UNSET,
        api_token_policy: Unsettable[JSONDictImmutable] = UNSET,
        api_token_version: Unsettable[int] = UNSET,
        avatar_urls: Unsettable[Mapping[str, str]] = UNSET,
        claim_username: bool = False,
        email: Unsettable[str] = UNSET,
        name: Unsettable[str] = UNSET,
        preferred_username: Unsettable[str | None] = UNSET,
        profile_version: Unsettable[int] = UNSET,
        service_account_id: Unsettable[str] = UNSET,
    ) -> None:
        """Initialize the service account.

        This can override any default attributes on the class, allowing
        extensions to either maintain service accounts fully or partly
        as subclasses or through instantiation.

        To claim a specific username (typically when one was already created
        and stored in the extension's settings), pass ``claim_username=True``
        and :samp:`preferred_username={username}`.

        Args:
            api_token_expiration_secs (int, optional):
                The expiration time for any created API tokens, in seconds.

                See :py:attr:`api_token_expiration`.

            api_token_policy (dict, optional):
                The token policy document to use for any created API tokens.

                See :py:attr:`api_token_policy`.

            api_token_version (dict, optional):
                The version of the API token state.

                See :py:attr:`api_token_version`.

            avatar_urls (dict, optional):
                Explicit avatars to use for managed accounts.

                See :py:attr:`avatar_urls`.

            claim_username (bool, optional):
                Whether to claim the preferred username as the managed
                service account user, if it already exists in the database.

                If set to ``True``, it's best to pass ``preferred_username``.

            email (str, optional):
                The e-mail address for the service account.

                See :py:attr:`email`.

            name (str, optional):
                The name of this service account.

                See :py:attr:`name`.

            preferred_username (str, optional):
                The preferred username for managed users.

                This should be provided if passing ``claim_username=True``.

                See :py:attr:`preferred_username``.

            profile_version (int, optional):
                The version of the profile state.

                See :py:attr:`profile_version`.

            service_account_id (str, optional):
                The unique ID of the service account.

                See :py:attr:`service_account_id`.

        Raises:
            ValueError:
                A required argument or attribute was not provided.
        """
        if service_account_id is not UNSET:
            self.service_account_id = service_account_id

        if not self.service_account_id:
            raise ValueError(
                f'{type(self).__name__}.service_account_id must be set or '
                f'provided during construction.'
            )

        if name is not UNSET:
            self.name = name

        if avatar_urls is not UNSET:
            self.avatar_urls = avatar_urls

        if email and email is not UNSET:
            self.email = email
        elif not self.email:
            siteconfig = SiteConfiguration.objects.get_current()
            noreply_email = siteconfig.get('mail_default_from')

            if noreply_email and isinstance(noreply_email, str):
                self.email = noreply_email

        if preferred_username and preferred_username is not UNSET:
            self.preferred_username = preferred_username
        elif not self.preferred_username:
            self.preferred_username = self.service_account_id

        if profile_version is not UNSET:
            self.profile_version = profile_version

        if api_token_expiration_secs is not UNSET:
            self.api_token_expiration_secs = api_token_expiration_secs

        if api_token_policy is not UNSET:
            self.api_token_policy = api_token_policy

        if api_token_version is not UNSET:
            self.api_token_version = api_token_version

        self._claim_username = claim_username
        self._user = None

    def get_user(
        self,
        *,
        _claim_map: (dict[str, ServiceAccount] | None) = None,
        _max_attempts: int = 100,
    ) -> User:
        """Return a user managed by this service account.

        If the user was already fetched from the database, it will be
        returned as-is. Otherwise, a new user will be looked up or created.
        This works as follows:

        1. If the service account is registered as claiming a specific
           username that already exists (such as one stored from a prior
           creation), that user will be returned.

        2. If one is not claimed, this will attempt to look up an existing
           user that's recorded as being associated with
           :py:attr:`service_account_id`.

        3. If the user was not found, one will be created.

        The username defined by :py:attr:`preferred_username` (or, if not set,
        :py:attr:`service_account_id`) will be used for lookup or creation.
        If a user is not being claimed by this service account, and another
        user already has this username, this will add a number to the username
        (starting with 1, then 2, etc.) until it finds a user that it can
        fetch or create.

        Args:
            _claim_map (dict, optional):
                A map for storing a claim for a username.

                This is meant only for internal use and is not considered
                part of the stable API.

            _max_attempts (int, optional):
                The default maximum number of attempts for claiming a user.

                This is meant only for internal use and is not considered
                part of the stable API.

        Returns:
            django.contrib.auth.models.User:
            The resulting user.

        Raises:
            reviewboard.accounts.errors.ServiceAccountUserError:
                A user could not be fetched, created, or claimed, and the
                maximum number of attempts have run out.
        """
        user = self._user

        if user is not None:
            return user

        # We'll set a default claim map if one isn't provided (the
        # non-registry case). Any username we may create will go in here
        # until we get a result. In the registry case, this will happen
        # within a thread lock, so there's no risk of two threads trying
        # to fetch or create the same user at the same time.
        if _claim_map is not None:
            claim_map = _claim_map
        else:
            claim_map = {}

        # If we're claiming the username, then we'll only need/want to
        # consider the explicit username given. If something goes wrong
        # here, we need to fail early instead of trying other variations,
        # otherwise state can get out of sync.
        #
        # Note that if the username has been claimed by multiple service
        # accounts, this is going to prevent the user from being fetched.
        preferred_username = self.preferred_username
        claim_username = self._claim_username

        if claim_username:
            if preferred_username in claim_map:
                # This username was already claimed somewhere else, but is
                # being claimed here. This is a fatal error.
                raise ServiceAccountUserError(
                    _(
                        'Cannot claim the service account username '
                        '"{username}". It has already been claimed by another '
                        'service account. This may be a conflict between two '
                        'extensions or a configuration error. Contact '
                        'support if you need assistance.'
                    ).format(username=preferred_username),
                )

            max_attempts = 1
        else:
            max_attempts = _max_attempts

        attempts = 0
        service_account_id = self.service_account_id
        base_username = preferred_username
        username = base_username
        created = False

        while attempts < max_attempts and user is None:
            attempts += 1

            # Try to fetch the user, creating if it doesn't exist. Django
            # will handle any IntegrityError checks from the creation and
            # attempt a re-fetch if needed.
            #
            # Note that unlike the service account code that existed in our
            # other implementations that preceded this, we're not creating an
            # object with the first/last names and e-mail populated, so we're
            # not going to be in a situation where the username matches but
            # the others don't, which would have triggered an IntegrityError.
            attempt_claim = False

            if username not in claim_map:
                attempt_claim = True

                try:
                    claim_map[username] = self
                    user, created = (
                        User.objects
                        .select_related('profile')
                        .get_or_create(username=username)
                    )
                except IntegrityError as e:
                    # This really should not happen, but we want to guard
                    # against it and try another variation (if we have
                    # attempts left).
                    user = None
                    created = False

                    logger.exception(
                        'Unexpected database integrity error fetching or '
                        'creating the service account user "%s". Trying '
                        'another variant. Error: %s',
                        username, e,
                    )

                if user:
                    # We got a user.
                    #
                    # This may be a user we explicitly want to claim (in which
                    # case we'll ensure the right state), a user matching the
                    # service ID (in which case we'll use it), or someone
                    # else's user (in which case we'll skip it).
                    if created or claim_username:
                        # We'll finish the loop and prepare this user.
                        break

                    # Check the implicit case.
                    if ((profile := user.get_profile(cached_only=True)) and
                        (service_account_id ==
                         profile.extra_data.get('service_account_id'))):
                        # This matches the service account, so finish the loop
                        # and prepare this user.
                        break

            # This was not a success. Remove from the claim map, clean
            # up state, and prepare for the next one.
            if attempt_claim:
                del claim_map[username]

            user = None
            created = False
            username = f'{base_username}-{attempts}'

        # We're out of the loop.
        #
        # At this point, we either have a user we can prepare and return,
        # or we've hit our max attempts.
        if user is None:
            # We've hit our max attempts. Log that.
            trace_id = str(uuid4())

            logger.error(
                '[%s] Failed to create a unique service account user for '
                '"%s". Tried %s variants, which were all taken. '
                'You may need to remove a user or contact support '
                'for help',
                trace_id, base_username, max_attempts,
            )

            raise ServiceAccountUserError(_(
                'Failed to create a unique service account user for '
                '"{username}" (error ID {error_id}). Please contact support.'
            ).format(error_id=trace_id,
                     username=base_username))

        # We have a user.
        if created:
            profile = Profile(user=user)
        else:
            profile = user.get_profile(cached_only=True) or Profile(user=user)

        profile_version = self.profile_version

        # If the user is newly-created, or has outdated state, then set
        # everything on the user.
        if (created or
            not profile.extra_data.get('service_account_id') or
            profile.extra_data.get('_profile_ver') != profile_version):
            # Create some default identifying information for the account.
            name = self.name
            email = self.email

            if name:
                # Default the name to the service account's name, but split
                # along spaces about half-way through to distribute across
                # the first/last name fields.
                name_parts = name.split(' ')

                if len(name_parts) > 1:
                    mid = len(name_parts) // 2
                    first_name = ' '.join(name_parts[:mid])
                    last_name = ' '.join(name_parts[mid:])
                else:
                    first_name = name
                    last_name = ''
            else:
                # Or if there's no name set, just use "Service Account".
                first_name = 'Service'
                last_name = 'Account'

            # Update the user to reflect that.
            user.email = email
            user.first_name = first_name
            user.last_name = last_name
            user.set_unusable_password()
            user.save(update_fields=(
                'email',
                'first_name',
                'last_name',
                'password',
            ))

            profile.should_send_email = False
            profile.extra_data['service_account_id'] = service_account_id
            profile.extra_data['_profile_ver'] = profile_version

            if profile.pk is None:
                profile.save()
                setattr(user, '_profile', profile)
            else:
                profile.save(update_fields=(
                    'extra_data',
                    'should_send_email',
                ))

            # If any avatars are set, associate them now.
            avatar_urls = self.avatar_urls

            if avatar_urls:
                avatar_service = avatar_services.get_avatar_service(
                    URLAvatarService.avatar_service_id,
                )
                avatar_service.setup(user, avatar_urls)

        self._user = user

        return user

    def get_api_token(
        self,
        *,
        local_site: LocalSite | None,
        min_validity_secs: int = 60 * 60 * 2,
    ) -> WebAPIToken:
        """Return an API token for the service account.

        If a matching token already exists, and will remain valid for a
        minimum amount of time (2 hours by default, if the configured
        expiration is more than that), then the token will be returned.

        If a valid token does not already exist, a new one will be created
        with policy document and expiration time provided on the service
        account.

        The API token will be keyed off from the :py:attr:`api_token_version`
        number.

        Note that this does not check if the expiration time or policy
        document has been modified to be different from those set on this
        service account. What's in the database will be returned as-is.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site the token must be tied to.

            min_validity_secs (int, optional):
                The minimum number of seconds left in any existing token
                considered for a result.

                This defaults to 2 hours.

        Returns:
            reviewboard.webapi.models.WebAPIToken:
            The resulting API token.
        """
        user = self.get_user()

        if expires_secs := self.api_token_expiration_secs:
            expires = timezone.now() + timedelta(seconds=expires_secs)

            if expires_secs < min_validity_secs:
                # The configured expiration time for the token is less than
                # the minimum time we're looking for, so ignore the minimum.
                min_validity_secs = 0
        else:
            expires = None

        return WebAPIToken.objects.get_or_create_client_token(
            client_name=(
                f'{self.service_account_id}_'
                f'v{self.api_token_version}'
            ),
            default_policy=self.api_token_policy,
            ignore_deprecated=True,
            expires=expires,
            min_validity_secs=min_validity_secs,
            local_site=local_site,
            user=user,
        )[0]


class ServiceAccountRegistry(Registry[ServiceAccount]):
    """A registry tracking service accounts.

    This tracks all service accounts registered via Review Board or
    extensions. Registered service accounts can only log in via API tokens,
    and do not take a seat on the license.

    Version Added:
        8.1
    """

    lookup_attrs = ('service_account_id',)

    default_errors = {
        **DEFAULT_ERRORS,
        ALREADY_REGISTERED: _(
            'Could not register service account %(item)s. It is already '
            'registered.'
        ),
        ATTRIBUTE_REGISTERED: _(
            'Could not register service account %(item)s: Another service '
            'account (%(duplicate)s) is already registered with the same ID.'
        ),
        NOT_REGISTERED: _(
            'No service account was found with the ID "%(attr_value)s".'
        ),
        UNREGISTER: _(
            'Could not unregister service account %(item)s: This service '
            'account was not yet registered.'
        ),
    }

    ######################
    # Instance variables #
    ######################

    #: A mapping of usernames to their service accounts.
    _by_username: dict[str, ServiceAccount]

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the registry.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent.

            **kwargs (tuple):
                Keyword arguments to pass to the parent.
        """
        self._by_username = {}

        super().__init__(*args, **kwargs)

    def get_for_service_account_id(
        self,
        service_account_id: str,
    ) -> ServiceAccount | None:
        """Return a service account with the given ID.

        Args:
            service_account_id (str):
                The ID of the service account.

        Returns:
            ServiceAccount:
            The registered service account, or ``None`` if not found.
        """
        return self.get('service_account_id', service_account_id)

    def get_for_username(
        self,
        username: str,
    ) -> ServiceAccount | None:
        """Return a service account for a given username.

        Args:
            username (str):
                The username to check for.

        Returns:
            ServiceAccount:
            The registered service account, or ``None`` if not found.
        """
        self.populate()

        return self._by_username.get(username)

    def on_item_registered(
        self,
        item: ServiceAccount,
    ) -> None:
        """Handle post-registration of a service account.

        This will fetch or create the managed user and add it to the
        username map.

        Args:
            item (ServiceAccount):
                The service account that was registered.
        """
        # Fetch or create the user, caching it in the process. This is being
        # done in this registration thread lock, preventing collisions
        try:
            item.get_user(_claim_map=self._by_username)
        except ServiceAccountUserError:
            return

    def on_item_unregistered(
        self,
        item: ServiceAccount,
    ) -> None:
        """Handle post-unregistration of a service account.

        This will remove its claim from the username map.

        Args:
            item (ServiceAccount):
                The service account that was unregistered.
        """
        # Remove the username claim from the map.
        if (user := item._user) is not None:
            self._by_username.pop(user.username, None)


#: The main service account registry for Review Board.
#:
#: Version Added:
#:     8.1
service_account_registry = ServiceAccountRegistry()



"""Upgrade custom overridden models from oauth2_provider from 0.9 to 1.6.3.

Version Added:
    5.0
"""

from django.db import models
from django.utils import timezone
from django_evolution.mutations import AddField, ChangeField


MUTATIONS = [
    AddField('Application', 'created', models.DateTimeField,
             initial=timezone.now),
    AddField('Application', 'updated', models.DateTimeField,
             initial=timezone.now),
    AddField('Application', 'algorithm', models.CharField,
             max_length=5, initial=''),
    ChangeField('Application', 'user', initial=None, null=True),
    ChangeField('Application', 'id', field_type=models.BigAutoField,
                primary_key=True),
]



"""Rename Profile.show_submitted to show_closed.

Version Added:
    2.0
"""

from django_evolution.mutations import RenameField


MUTATIONS = [
    RenameField('Profile', 'show_submitted', 'show_closed'),
]



import pymdownx.emoji
from bleach.sanitizer import Cleaner
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model
from django.utils.encoding import force_str
from django.utils.html import escape
from djblets import markdown as djblets_markdown
from djblets.siteconfig.models import SiteConfiguration
from markdown import markdown


# Keyword arguments used when calling a Markdown renderer function.
#
# We use XHTML instead of HTML5 to ensure the results can be parsed by an
# XML parser, needed for doing diffs in change descriptions and the Markdown
# review UI.
MARKDOWN_KWARGS = {
    'enable_attributes': False,
    'output_format': 'xhtml',
    'lazy_ol': False,
    'extensions': [
        'markdown.extensions.fenced_code',
        'markdown.extensions.codehilite',
        'markdown.extensions.sane_lists',
        'markdown.extensions.tables',
        'markdown.extensions.nl2br',
        'pymdownx.tilde',
        'pymdownx.emoji',
        'djblets.markdown.extensions.escape_html',
        'djblets.markdown.extensions.wysiwyg',
    ],
    'extension_configs': {
        'markdown.extensions.codehilite': {
            'guess_lang': False,
            'linenums': False,
        },
        'pymdownx.emoji': {
            'emoji_index': pymdownx.emoji.gemoji,
            'options': {
                'classes': 'emoji',
                'image_path': ('https://github.githubassets.com/images/icons/'
                               'emoji/unicode/'),
                'non_standard_image_path': ('https://github.githubassets.com/'
                                            'images/icons/emoji/'),
            },
        },
    },
}


#: A list of HTML tags considered to be safe in Markdown-generated output.
#:
#: Anything not in this list will be escaped when sanitizing the resulting
#: HTML.
#:
#: Version Added:
#:     3.0.22
SAFE_MARKDOWN_TAGS = [
    'a',
    'b',
    'blockquote',
    'br',
    'code',
    'dd',
    'del',
    'div',
    'dt',
    'em',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'hr',
    'i',
    'img',
    'li',
    'ol',
    'p',
    'pre',
    'span',
    'strong',
    'sub',
    'sup',
    'table',
    'tbody',
    'td',
    'foot',
    'th',
    'thead',
    'tr',
    'tt',
    'ul',
]


#: Mappings of HTML tags to attributes considered to be safe for Markdown.
#:
#: Anything not in this list will be removed when sanitizing the resulting
#: HTML.
#:
#: Version Added:
#:     3.0.22
SAFE_MARKDOWN_ATTRS = {
    '*': ['class', 'id'],
    'a': ['href', 'alt', 'title'],
    'img': ['src', 'alt', 'title'],
}


#: A list of protocols considered safe for URLs.
#:
#: This can be overridden by setting
#: ``settings.ALLOWED_MARKDOWN_URL_PROTOCOLS``.
#:
#: Version Added:
#:     3.0.24
SAFE_MARKDOWN_URL_PROTOCOLS = ['http', 'https', 'mailto']


def markdown_escape_field(obj, field_name):
    """Escapes Markdown text in a model or dictionary's field.

    This is a convenience around markdown_escape to escape the contents of
    a particular field in a model or dictionary.
    """
    if isinstance(obj, Model):
        setattr(obj, field_name,
                djblets_markdown.markdown_escape(getattr(obj, field_name)))
    elif isinstance(obj, dict):
        obj[field_name] = djblets_markdown.markdown_escape(obj[field_name])
    else:
        raise TypeError('Unexpected type %r passed to markdown_escape_field'
                        % obj)


def markdown_unescape_field(obj, field_name):
    """Unescapes Markdown text in a model or dictionary's field.

    This is a convenience around markdown_unescape to unescape the contents of
    a particular field in a model or dictionary.
    """
    if isinstance(obj, Model):
        setattr(obj, field_name,
                djblets_markdown.markdown_unescape(getattr(obj, field_name)))
    elif isinstance(obj, dict):
        obj[field_name] = djblets_markdown.markdown_unescape(obj[field_name])
    else:
        raise TypeError('Unexpected type %r passed to markdown_unescape_field'
                        % obj)


def normalize_text_for_edit(user, text, rich_text, escape_html=True):
    """Normalizes text, converting it for editing.

    This will normalize text for editing based on the rich_text flag and
    the user settings.

    If the text is not in Markdown and the user edits in Markdown by default,
    this will return the text escaped for edit. Otherwise, the text is
    returned as-is.
    """
    if text is None:
        return ''

    if not rich_text and is_rich_text_default_for_user(user):
        # This isn't rich text, but it's going to be edited as rich text,
        # so escape it.
        text = djblets_markdown.markdown_escape(text)

    if escape_html:
        text = escape(text)

    return text


def markdown_render_conditional(text, rich_text):
    """Return the escaped HTML content based on the rich_text flag."""
    if rich_text:
        return render_markdown(text)
    else:
        return escape(text)


def is_rich_text_default_for_user(user):
    """Returns whether the user edits in Markdown by default."""
    if user.is_authenticated:
        try:
            return user.get_profile().should_use_rich_text
        except ObjectDoesNotExist:
            pass

    siteconfig = SiteConfiguration.objects.get_current()

    return siteconfig.get('default_use_rich_text')


def markdown_set_field_escaped(obj, field, escaped):
    """Escapes or unescapes the specified field in a model or dictionary."""
    if escaped:
        markdown_escape_field(obj, field)
    else:
        markdown_unescape_field(obj, field)


def clean_markdown_html(html):
    """Return a cleaned, secure version of Markdown-rendered HTML/XHTML.

    This will sanitize Markdown-rendered HTML, ensuring that only a trusted
    list of HTML tags, attributes, and URI schemes are included in the
    HTML. Anything else will be left out or transformed into a safe
    representation of the original content.

    The result will always be in XHTML form, to allow for XML processing of the
    content.

    Version Added:
        3.0.24

    Args:
        html (unicode):
            The Markdown-rendered HTML to clean.

    Returns:
        unicode:
        A sanitizied XHTML representation of the Markdown-rendered HTML.
    """
    # Allow users to override the protocols. We're checking for this
    # dynamically, partly to ease unit testing, and partly to eventually
    # allow dynamic configuration.
    safe_url_protocols = SAFE_MARKDOWN_URL_PROTOCOLS
    custom_safe_url_protocols = settings.ALLOWED_MARKDOWN_URL_PROTOCOLS

    if custom_safe_url_protocols:
        safe_url_protocols = (set(safe_url_protocols) |
                              set(custom_safe_url_protocols))

    # Create a bleach HTML cleaner, and override settings on the html5lib
    # serializer it contains to ensure we use self-closing HTML tags, like
    # <br/>. This is needed so that we can parse the resulting HTML in
    # Djblets for things like Markdown diffing.
    cleaner = Cleaner(tags=SAFE_MARKDOWN_TAGS,
                      attributes=SAFE_MARKDOWN_ATTRS,
                      protocols=safe_url_protocols)
    cleaner.serializer.use_trailing_solidus = True

    return cleaner.clean(html)


def render_markdown(text):
    """Render Markdown text to XHTML.

    The Markdown text will be sanitized to prevent injecting custom HTML
    or dangerous links. It will also enable a few plugins for code
    highlighting and sane lists.

    It's rendered to XHTML in order to allow the element tree to be easily
    parsed for code review and change description diffing.

    Args:
        text (bytes or unicode):
            The Markdown text to render.

            If this is a byte string, it must represent UTF-8-encoded text.

    Returns:
        unicode:
        The Markdown-rendered XHTML.
    """
    return clean_markdown_html(markdown(force_str(text), **MARKDOWN_KWARGS))


def render_markdown_from_file(f):
    """Render Markdown text from a file to XHTML.

    The Markdown text will be sanitized to prevent injecting custom HTML.
    It will also enable a few plugins for code highlighting and sane lists.

    Version Changed:
        3.0.24:
        This has been updated to sanitize the rendered HTML to avoid any
        security issues.

    Args:
        f (file):
            The file stream to read from.

    Returns:
        unicode:
        The Markdown-rendered XHTML.
    """
    return clean_markdown_html(djblets_markdown.render_markdown_from_file(
        f, **MARKDOWN_KWARGS))



import logging

from django.db.models.signals import post_init

from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


def _migrate_scmtool_ids(instance, **kwargs):
    """Migrate SCMTool IDs on repositories, if missing.

    Starting in Review Board 5.0, SCMTools began to be tracked by a
    :py:attr:`~reviewboard.scmtools.models.Repository.scmtool_id` attribute.
    This is efficiently set when upgrading the database, but if any fail to
    migrate (such as if a package-provided SCMTool can't be loaded during
    upgrade) or if data is injected/imported into the database after upgrade,
    this may be NULL.

    This handler will detect this whenever a repository is instantiated, and
    will try to perform a migration automatically.

    This will only attempt the SCMTool ID migration if working with a full
    repository without any deferred fields, in order to avoid any unwanted
    database access.

    Args:
        instance (reviewboard.scmtools.models.Repository):
            The repository instance being initialized.

        **kwargs (dict, unused):
            Additional keyword arguments.
    """
    if instance.get_deferred_fields():
        # We don't want to risk any extra queries here. This can be called
        # during the pre-upgrade steps, and an unwanted query can be fatal
        # to the upgrade process. So if there are deferred fields, just
        # skip this handler.
        return

    if (instance.scmtool_id is None and
        instance.tool_id is not None):
        # This wasn't set during upgrade. The package/extension may have
        # been missing at that time. We'll attempt to find it now and
        # upgrade the repository.
        try:
            tool = instance.tool
            scmtool_id = tool.scmtool_id
        except Exception as e:
            logger.error('Attempted to upgrade the SCMTool state for '
                         'repository ID %s, but the SCMTool "%s" could '
                         'not be loaded: %s',
                         instance.pk, tool.name, e)
            return

        if scmtool_id:
            instance.scmtool_id = scmtool_id

            if instance.pk:
                try:
                    instance.save(update_fields=('scmtool_id',))
                except Exception as e:
                    logger.error('Unable to save migrated scmtool_id "%s" '
                                 'for repository ID %s: %s',
                                 scmtool_id, instance.pk, e)


def connect_signal_handlers():
    """Connect SCMTool-related signal handlers.

    Version Added:
        5.0
    """
    post_init.connect(_migrate_scmtool_ids, sender=Repository)



"""Unit tests for reviewboard.admin.forms.SSHSettingsForm."""

import os
import shutil
import tempfile

from django.utils.encoding import force_str

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.ssh.client import SSHClient
from reviewboard.testing.testcase import TestCase


class SSHSettingsFormTestCase(TestCase):
    """Unit tests for reviewboard.admin.forms.SSHSettingsForm."""

    fixtures = ['test_users']

    def setUp(self):
        """Set up this test case."""
        super(SSHSettingsFormTestCase, self).setUp()

        # Setup temp directory to prevent the original ssh related
        # configurations been overwritten.
        self.old_home = os.getenv('HOME')
        self.tempdir = tempfile.mkdtemp(prefix='rb-tests-home-')
        os.environ[str('RBSSH_ALLOW_AGENT')] = str('0')
        self._set_home(self.tempdir)

        self.ssh_client = SSHClient()

    def tearDown(self):
        """Tear down this test case."""
        super(SSHSettingsFormTestCase, self).tearDown()

        self._set_home(self.old_home)

        if self.tempdir:
            shutil.rmtree(self.tempdir)

    def _set_home(self, homedir):
        """Set the $HOME environment variable."""
        os.environ[str('HOME')] = force_str(homedir)

    def test_generate_key(self):
        """Testing SSHSettingsForm POST with generate_key=1"""
        # Should have no ssh key at this point.
        self.assertIsNone(self.ssh_client.get_user_key())

        # Send post request with 'generate_key' = 1.
        self.client.login(username='admin', password='admin')
        response = self.client.post(local_site_reverse('settings-ssh'), {
            'generate_key': 1,
        })

        # On success, the form returns HTTP 302 (redirect).
        self.assertEqual(response.status_code, 302)

        # Check whether the key has been created.
        self.assertIsNotNone(self.ssh_client.get_user_key())

    def test_delete_key(self):
        """Testing SSHSettingsForm POST with delete_key=1"""
        # Should have no ssh key at this point, generate one.
        self.assertIsNone(self.ssh_client.get_user_key())
        self.ssh_client.generate_user_key()
        self.assertIsNotNone(self.ssh_client.get_user_key())

        # Send post request with 'delete_key' = 1.
        self.client.login(username='admin', password='admin')
        response = self.client.post(local_site_reverse('settings-ssh'), {
            'delete_key': 1,
        })

        # On success, the form returns HTTP 302 (redirect).
        self.assertEqual(response.status_code, 302)

        # Check whether the key has been deleted.
        self.assertIsNone(self.ssh_client.get_user_key())



"""Registry for actions.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
from collections import defaultdict
from typing import Iterator

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         NOT_REGISTERED,
                                         RegistryState)
from housekeeping import func_deprecated

from reviewboard.actions.base import (ActionAttachmentPoint,
                                      ActionPlacement,
                                      AttachmentPoint,
                                      BaseAction,
                                      BaseGroupAction)
from reviewboard.actions.renderers import ButtonActionRenderer
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.registries.registry import OrderedRegistry, Registry


logger = logging.getLogger(__name__)


class ActionAttachmentPointsRegistry(Registry[ActionAttachmentPoint]):
    """A registry for action attachment points.

    Version Added:
        8.0
    """

    lookup_attrs = ['attachment_point_id']

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered attachment point.'
        ),
        NOT_REGISTERED: _(
            '"%(attr_value)s" is not a registered attachment point.'
        ),
    }

    def get_defaults(self) -> Iterator[ActionAttachmentPoint]:
        """Yield the built-in attachment points.

        Yields:
            reviewboard.actions.base.BaseAction:
            The built-in attachment points.
        """
        from reviewboard.admin.actions import (
            get_default_admin_attachment_points,
        )

        yield from get_default_admin_attachment_points()
        yield from (
            ActionAttachmentPoint(AttachmentPoint.NON_UI),
            ActionAttachmentPoint(AttachmentPoint.HEADER),
            ActionAttachmentPoint(AttachmentPoint.REVIEW_REQUEST_LEFT),
            ActionAttachmentPoint(AttachmentPoint.REVIEW_REQUEST),
            ActionAttachmentPoint(AttachmentPoint.UNIFIED_BANNER),
            ActionAttachmentPoint(
                AttachmentPoint.QUICK_ACCESS,
                default_action_renderer_cls=ButtonActionRenderer,
            ),
        )

    def get_attachment_point(
        self,
        attachment_point_id: str,
    ) -> ActionAttachmentPoint | None:
        """Return the attachment point with the given ID.

        Args:
            attachment_point_id (str):
                The attachment point ID to look up.

        Returns:
            reviewboard.actions.base.ActionAttachmentPoint:
            The resulting attachment point instance, or ``None`` if not found.
        """
        return self.get('attachment_point_id', attachment_point_id)


class ActionsRegistry(OrderedRegistry[BaseAction]):
    """A registry for actions.

    Version Added:
        6.0
    """

    lookup_attrs = ['action_id']

    errors = {
        ALREADY_REGISTERED: _('"%(item)s" is already a registered action.'),
        NOT_REGISTERED: _('"%(attr_value)s" is not a registered action.'),
    }

    MAX_DEPTH = 2

    ######################
    # Instance variables #
    ######################

    #: Action lookup by attachment point.
    #:
    #: Version Added:
    #:     8.0
    _by_attachment_point: defaultdict[
        str,
        dict[
            str,
            tuple[BaseAction, ActionPlacement]
        ]
    ]

    #: Placements deferring registration until a dependency is registered.
    #:
    #: Keys are tuples in the form of ``(parent_id, attachment)``.
    #
    #: Values are dictionaries are in the form of ``{action_id: action}``.
    #:
    #: Version Added:
    #:     8.0
    _deferred_placements: defaultdict[
        tuple[str, str],
        dict[str, BaseAction]
    ]

    def __init__(self) -> None:
        """Initialize the registry."""
        super().__init__()

        self._by_attachment_point = defaultdict(dict)
        self._deferred_placements = defaultdict(dict)

    def get_defaults(self) -> Iterator[BaseAction]:
        """Yield the built-in actions.

        Yields:
            reviewboard.actions.base.BaseAction:
            The built-in actions.
        """
        from reviewboard.accounts.actions import (
            AccountMenuAction,
            AdminAction,
            DocumentationAction,
            FollowFacebookAction,
            FollowBlueSkyAction,
            FollowLinkedInAction,
            FollowMastodonAction,
            FollowMenuAction,
            FollowRedditAction,
            FollowNewsAction,
            FollowTwitterAction,
            FollowYouTubeAction,
            LoginAction,
            LogoutAction,
            MyAccountAction,
            SupportAction,
            SupportMenuAction,
        )
        from reviewboard.admin.actions import get_default_admin_actions
        from reviewboard.reviews.actions import (
            AddGeneralCommentAction,
            ArchiveAction,
            ArchiveMenuAction,
            CloseMenuAction,
            CloseCompletedAction,
            CloseDiscardedAction,
            CreateReviewAction,
            DeleteAction,
            DownloadDiffAction,
            EditReviewAction,
            LegacyAddGeneralCommentAction,
            LegacyEditReviewAction,
            LegacyShipItAction,
            MuteAction,
            ReviewMenuAction,
            ShipItAction,
            StarAction,
            UpdateMenuAction,
            UploadDiffAction,
            UploadFileAction,
        )

        # The order here is important, and will reflect the order that items
        # appear in the UI.
        yield from (
            # Header bar
            AccountMenuAction(),
            LoginAction(),
            MyAccountAction(),
            AdminAction(),
            LogoutAction(),
            SupportMenuAction(),
            DocumentationAction(),
            SupportAction(),
            FollowMenuAction(),
            FollowNewsAction(),
            FollowBlueSkyAction(),
            FollowFacebookAction(),
            FollowLinkedInAction(),
            FollowMastodonAction(),
            FollowRedditAction(),
            FollowTwitterAction(),
            FollowYouTubeAction(),

            # Review request actions (left side)
            StarAction(),
            ArchiveMenuAction(),
            ArchiveAction(),
            MuteAction(),

            # Review request actions (right side)
            CloseMenuAction(),
            CloseCompletedAction(),
            CloseDiscardedAction(),
            DeleteAction(),
            UpdateMenuAction(),
            UploadDiffAction(),
            UploadFileAction(),
            DownloadDiffAction(),
            LegacyEditReviewAction(),
            LegacyShipItAction(),
            LegacyAddGeneralCommentAction(),

            # Unified banner actions
            ReviewMenuAction(),
            CreateReviewAction(),
            EditReviewAction(),
            AddGeneralCommentAction(),
            ShipItAction(),
        )

        yield from get_default_admin_actions()

    def on_reset(self) -> None:
        """Handle cleanup after resetting the registry.

        This clears cached state used for looking up actions by attachment
        point and deferred placements.
        """
        super().on_reset()

        # If everything went as expected, these lists should be empty. But
        # on the off-chance that there's a bug, we want to assert in debug
        # mode so we can catch this without affecting production.
        if settings.DEBUG or settings.RUNNING_TEST:
            assert self._by_attachment_point == {}, self._by_attachment_point
            assert self._deferred_placements == {}, self._deferred_placements

        self._by_attachment_point.clear()
        self._deferred_placements.clear()

    def on_item_registered(
        self,
        action: BaseAction,
        /,
    ) -> None:
        """Handle post-registration placement of an action.

        This will process any placements for the action, putting them in
        the right place within an attachment point's action hierarchy.

        It will also handle deferring any placements for parent actions not
        yet registered, and processing of deferred placements for actions
        now registered.

        Args:
            action (reviewboard.actions.base.BaseAction):
                The action that was registered.
        """
        super().on_item_registered(action)

        action.parent_registry = self

        action_id = action.action_id
        assert action_id is not None

        # Sanity-check each placement to make sure it has a valid parent
        # reference (if needed) and doesn't exceed the depth limit. Any that
        # does will be excluded from the page, but any that do will remain.
        #
        # In any case, the action will still be registered. Note that this is
        # a change from Review Board 6/7, since actions can now be registered
        # and referenced from multiple locations. It's no longer fatal for
        # one of these checks to fail.
        deferred_placements = self._deferred_placements
        seen_attachments: set[str] = set()

        for placement in (action.placements or []):
            attachment = placement.attachment

            if attachment in seen_attachments:
                logger.warning(
                    'Action "%s" has been placed in attachment "%s" more '
                    'than once. Only the first instance will be placed.',
                    action_id, attachment)
                continue

            seen_attachments.add(attachment)

            parent_id = placement.parent_id
            parent_placement: (ActionPlacement | None) = None
            parent: (BaseAction | None) = None

            if parent_id:
                parent = self.get_action(parent_id)

                if parent is None:
                    # Old versions of the extension hooks for actions
                    # encouraged people to register their child actions before
                    # the parent. Furthermore, since actions can be placed in
                    # multiple parents, not all parents may be registered.
                    # Defer these placements until they've been set up.
                    logger.debug(
                        'Deferring placement of action "%s" in attachment '
                        'point "%s" because parent action "%s" is not yet '
                        'registered.',
                        action_id, attachment, parent_id)

                    defer_key = (parent_id, attachment)
                    deferred_placements[defer_key][action_id] = action
                    continue

                try:
                    parent_placement = parent.get_placement(attachment)
                except KeyError as e:
                    logger.error(
                        'Invalid placement %r found while registering action '
                        '%r in parent %r: %s',
                        attachment, action_id, parent_id, e)
                    continue

                if parent_placement is None:
                    logger.error(
                        'Action "%s" refers to a parent action "%s" in '
                        'attachment point "%s", but the parent is not '
                        'placed in that attachment point. Skipping.',
                        action_id, parent.action_id, attachment)
                    continue
                elif parent_placement.depth + 1 > self.MAX_DEPTH:
                    logger.error(
                        'Action "%s" exceeds the maximum depth limit of %s '
                        'for attachment point(s) "%s".',
                        action_id, self.MAX_DEPTH, attachment)
                    continue

            self._setup_placement(
                action=action,
                placement=placement,
                parent=parent,
                parent_placement=parent_placement,
            )

    def on_item_unregistering(
        self,
        action: BaseAction,
        /,
    ) -> None:
        """Handle pre-unregistration tasks for an action.

        Args:
            action (reviewboard.actions.base.BaseAction):
                The action to unregister.
        """
        for placement in (action.placements or []):
            self._remove_placement(action, placement,
                                   unregistered=True)

        action.parent_registry = None

    def get_action(
        self,
        action_id: str,
    ) -> BaseAction | None:
        """Return the action with the given ID.

        Version Added:
            8.0

        Args:
            action_id (str):
                The action ID to look up.

        Returns:
            reviewboard.actions.base.BaseAction:
            The resulting action instance, or ``None`` if not found.
        """
        return self.get('action_id', action_id)

    def get_for_attachment(
        self,
        attachment: str,
        *,
        include_children: bool = False,
    ) -> Iterator[BaseAction]:
        """Yield actions for the given attachment point.

        Args:
            attachment (reviewboard.actions.base.AttachmentPoint):
                The attachment point.

            include_children (bool, optional):
                Whether to also include children of menus. If ``False``, this
                will only yield the top-level items.

        Yields:
            reviewboard.actions.base.BaseAction:
            The actions for the given attachment point.
        """
        self.populate()

        attachment_actions = self._by_attachment_point.get(attachment)

        if attachment_actions:
            for action, placement in attachment_actions.values():
                if placement.parent_id is None or include_children:
                    yield action

    @func_deprecated(RemovedInReviewBoard10_0Warning, message=(
        '%(func_name)s is deprecated and support will be removed in '
        '%(product)s %(version)s. Please use '
        'action.get_placement(...).child_actions instead.'
    ))
    def get_children(
        self,
        parent_id: str,
    ) -> Iterator[BaseAction]:
        """Return the children for a menu action.

        Args:
            parent_id (str):
                The ID of the parent.

        Yields:
            reviewboard.actions.base.BaseAction:
            The actions that are contained within the menu.
        """
        parent = self.get_action(parent_id)

        assert parent is not None, (
            f'Action {parent_id!r} was not registered when calling '
            f'get_children().'
        )

        yield from parent.child_actions

    def _setup_placement(
        self,
        *,
        action: BaseAction,
        placement: ActionPlacement,
        parent: BaseAction | None,
        parent_placement: ActionPlacement | None,
    ) -> None:
        """Set up a placement for an action.

        This will place the action within the attachment, parented to another
        action if specified.

        If the action being placed was a blocker for any deferred placements,
        those placements will be processed.

        Args:
            action (reviewboard.actions.base.BaseAction):
                The action being placed.

            placement (reviewboard.actions.base.ActionPlacement):
                The placement for this action.

            parent (reviewboard.actions.base.BaseAction):
                The parent action this will be placed under, if any.

            parent_placement (reviewboard.actions.base.ActionPlacement):
                The placement for the parent action, if a parent is specified.
        """
        attachment = placement.attachment
        action_id = action.action_id

        # Store by attachment points, for quick lookup.
        self._by_attachment_point[attachment][action_id] = (action, placement)

        # Establish the parent/child relationship between actions.
        if parent is not None:
            assert isinstance(parent, BaseGroupAction)
            assert parent_placement is not None

            placement.parent_action = parent
            parent_placement.child_actions.append(action)
        else:
            placement.parent_action = None

        # Go through any deferred placements parented here and add them,
        # recursively.
        deferred = self._deferred_placements.pop((action_id, attachment), {})

        for deferred_action in deferred.values():
            if self.state != RegistryState.POPULATING:
                # If we're not initially populating the registry with the
                # defaults shipped with Review Board, leave a debug message
                # that this will be deferred in case an extension author is
                # unsure why their action isn't showing up.
                logger.debug(
                    'Adding deferred placement of action "%s" in attachment '
                    'point "%s" for parent action "%s".',
                    deferred_action.action_id, attachment,
                    action_id)

            self._setup_placement(
                action=deferred_action,
                placement=deferred_action.get_placement(attachment),
                parent=action,
                parent_placement=placement,
            )

    def _remove_placement(
        self,
        action: BaseAction,
        placement: ActionPlacement,
        *,
        unregistered: bool = False,
    ) -> None:
        """Remove a placement for an action.

        This will remove a placement as part of either unregistering an
        action or from removing or re-deferring a parent placement.

        The placement will be removed from the parent placement and from
        lookup tables. All child actions in the placement will be removed.

        Removed actions are put back into a deferred state so they can be
        placed again later if the parent placement is re-added.

        Args:
            action (reviewboard.actions.base.BaseAction):
                The action for the placement.

            placement (reviewboard.actions.base.ActionPlacement):
                The placement being removed.

            unregistered (bool, optional):
                Whether the action is being explicitly unregistered.
        """
        action_id = action.action_id
        by_attachment_point = self._by_attachment_point

        attachment = placement.attachment
        attachment_actions = by_attachment_point[attachment]
        deferred_placements = self._deferred_placements

        # Remove placements for any children that depend on this placement
        # so they can be re-added if this action is registered again.
        if placement.child_actions:
            for child_action in list(placement.child_actions):
                child_placement = child_action.get_placement(attachment)

                if child_placement:
                    self._remove_placement(child_action, child_placement)
        else:
            # If nothing depends on this placement, we can remove it from
            # the deferment list if present.
            deferred_placements.pop((action_id, attachment), None)

        # If this placement is parented to another (whether it currently
        # exists in the registry or not), it may need to be deferred or
        # removed from a deferred list.
        if (parent_id := placement.parent_id):
            defer_key = (parent_id, attachment)

            if not unregistered:
                deferred_placements[defer_key][action_id] = action
            elif not placement.child_actions:
                # This action is being unregistered, and it has a parent but
                # no child actions. This means it can be safely removed from
                # any deferred list, since adding a parent action back to
                # the registry would not need to place this again.
                if defer_key in deferred_placements:
                    deferred_placements[defer_key].pop(action_id, None)

                    if not deferred_placements[defer_key]:
                        # There are no more deferred placements for this
                        # parent and attachment, so delete its entry.
                        del deferred_placements[defer_key]

        # If this placement is parented to another, remove it from that
        # placement's list of children.
        if (parent_action := placement.parent_action):
            parent_placement = parent_action.get_placement(attachment)

            if parent_placement is not None:
                parent_placement.child_actions.remove(action)

            placement.parent_action = None

        # Remove this from the per-attachment lookups.
        #
        # Note that it may not be in this list, because we remove this both
        # when processing children of an unregistered action and when
        # processing the unregistered action itself. We can end up here
        # multiple times when tearing down multiple actions, such as during
        # extension shutdown or registry reset.
        attachment_actions.pop(action_id, None)

        # Clear out this attachment point lookup if there are no longer
        # any actions left.
        if not attachment_actions:
            del by_attachment_point[attachment]



"""Hosting service support."""

from __future__ import annotations

from django.dispatch import receiver

from reviewboard.signals import initializing


@receiver(initializing, dispatch_uid='populate_hosting_services')
def _populate_registry(**kwargs) -> None:
    """Populate the hosting services registry.

    This will pre-populate the hosting services registry, ensuring
    that any URLs are automatically added and available for use.

    Version Added:
        6.0

    Args:
        **kwargs (dict, unused):
            Keyword arguments sent by the signal.
    """
    # We populate this here because we need to ensure any URLs are registered
    # once Review Board has initialized. We also need to ensure this is done
    # during unit test runs when the environment is re-initialized. It's not
    # sufficient to do this during AppConfig.ready.
    from reviewboard.hostingsvcs.base import hosting_service_registry

    hosting_service_registry.populate()



"""Add FileDiffData.extra_data and remove delete_count and insert_count.

Version Added:
    2.0
"""

from django_evolution.mutations import AddField, DeleteField
from djblets.db.fields import JSONField


MUTATIONS = [
    AddField('FileDiffData', 'extra_data', JSONField, null=True),
    DeleteField('FileDiffData', 'insert_count'),
    DeleteField('FileDiffData', 'delete_count'),
]



"""Unit tests for the ServerInfoResource API."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Tuple

from django.conf import settings
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard import get_version_string, get_package_version, is_release
from reviewboard.admin.server import get_server_url
from reviewboard.reviews.ui.base import ReviewUI, register_ui, unregister_ui
from reviewboard.webapi.resources import resources
from reviewboard.webapi.server_info import get_capabilities
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import server_info_mimetype
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import get_server_info_url

if TYPE_CHECKING:
    from django.contrib.auth.models import User


class ResourceTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the ServerInfoResource APIs."""

    fixtures = ['test_users']
    sample_api_url = 'info/'
    resource = resources.server_info

    def setup_http_not_allowed_list_test(self, user):
        return get_server_info_url()

    def setup_http_not_allowed_item_test(self, user):
        return get_server_info_url()

    def compare_item(self, item_rsp, obj):
        self.assertIn('product', item_rsp)
        self.assertIn('site', item_rsp)
        self.assertIn('capabilities', item_rsp)

        product_rsp = item_rsp['product']
        self.assertEqual(product_rsp['name'], 'Review Board')
        self.assertEqual(product_rsp['version'], get_version_string())
        self.assertEqual(product_rsp['package_version'], get_package_version())
        self.assertEqual(product_rsp['is_release'], is_release())

        site_rsp = item_rsp['site']
        self.assertTrue(site_rsp['url'].startswith(get_server_url()))
        self.assertEqual(site_rsp['administrators'], [
            {
                'name': name,
                'email': email,
            }
            for name, email in settings.ADMINS
        ])
        self.assertEqual(site_rsp['time_zone'], settings.TIME_ZONE)

        self.assertEqual(item_rsp['capabilities'], get_capabilities())

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(
        self,
        user: User,
        with_local_site: bool,
        local_site_name: Optional[str],
    ) -> Tuple[str, str, Any]:
        """Set up a basic HTTP GET unit test.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the API requests.

            with_local_site (bool):
                Whether the test is being performed on a Local Site.

            local_site_name (str or None):
                The name of the Local Site to test against.

                This will be ``None`` if testing against the global site.

        Returns:
            tuple:
            A 3-tuple of:

            Tuple:
                0 (str):
                    The URL to the API resource to access.

                1 (str):
                    The expected mimetype of the response.

                2 (object):
                    The item to compare to in :py:meth:`compare_item`.
        """
        return (get_server_info_url(local_site_name),
                server_info_mimetype,
                None)

    @webapi_test_template
    def test_get_with_review_ui_mimetype(self) -> None:
        """Testing the GET <URL> API with a Review UI adding to mimetypes in
        capabilities
        """
        url, mimetype, obj = self.setup_basic_get_test(self.user, False, None)

        class TestReviewUI(ReviewUI):
            supported_mimetypes = ['application/rbtest']

        register_ui(TestReviewUI)

        try:
            rsp = self.api_get(url, expected_mimetype=mimetype)
            assert rsp is not None

            capabilities = rsp['info']['capabilities']

            self.assertIn('application/rbtest',
                          capabilities['review_uis']['supported_mimetypes'])
        finally:
            unregister_ui(TestReviewUI)

    @webapi_test_template
    def test_get_registered_scmtools(self) -> None:
        """Testing the GET <URL> API registered SCMTools"""
        url, mimetype, obj = self.setup_basic_get_test(self.user, False, None)

        rsp = self.api_get(url, expected_mimetype=mimetype)
        assert rsp is not None

        capabilities = rsp['info']['capabilities']

        self.assertIn('git',
                      capabilities['scmtools']['supported_tools'])



"""Unit tests for reviewboard.scmtools.crypto_utils."""

from django.test.utils import override_settings

from reviewboard.scmtools.crypto_utils import (aes_decrypt,
                                               aes_encrypt,
                                               decrypt_password,
                                               encrypt_password,
                                               get_default_aes_encryption_key)
from reviewboard.testing.testcase import TestCase


@override_settings(SECRET_KEY='abcdefghijklmnopqrstuvwxyz012345')
class CryptoUtilsTests(TestCase):
    """Unit tests for reviewboard.scmtools.crypto_utils."""

    PLAIN_TEXT_UNICODE = 'this is a test 123 ^&*'
    PLAIN_TEXT_BYTES = b'this is a test 123 ^&*'

    PASSWORD_UNICODE = 'this is a test 123 ^&*'
    PASSWORD_BYTES = b'this is a test 123 ^&*'

    CUSTOM_KEY = b'0123456789abcdef'

    def test_aes_decrypt(self):
        """Testing aes_decrypt"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = (
            b'\xfb\xdc\xb5h\x15\xa1\xb2\xdc\xec\xf1\x14\xa9\xc6\xab\xb2J\x10'
            b'\'\xd4\xf6&\xd4k9\x82\xf6\xb5\x8bmu\xc8E\x9c\xac\xc5\x04@B'
        )

        decrypted = aes_decrypt(encrypted)
        self.assertIsInstance(decrypted, bytes)
        self.assertEqual(decrypted, self.PLAIN_TEXT_BYTES)

    def test_aes_decrypt_with_unicode(self):
        """Testing aes_decrypt with Unicode data"""
        expected_message = ('The data to decrypt must be of type "bytes", '
                            'not "<class \'str\'>"')

        with self.assertRaisesMessage(TypeError, expected_message):
            aes_decrypt('abc')

    def test_aes_decrypt_with_custom_key(self):
        """Testing aes_decrypt with custom key"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = (
            b'\x9cd$e\xb1\x9e\xe0z\xb8[\x9e!\xf2h\x90\x8d\x82f%G4\xc2\xf0'
            b'\xda\x8dr\x81ER?S6\x12%7\x98\x89\x90'
        )

        decrypted = aes_decrypt(encrypted, key=self.CUSTOM_KEY)
        self.assertIsInstance(decrypted, bytes)
        self.assertEqual(decrypted, self.PLAIN_TEXT_BYTES)

    def test_aes_decrypt_with_custom_key_unicode(self):
        """Testing aes_decrypt with custom key as Unicode"""
        encrypted = (
            b'\x9cd$e\xb1\x9e\xe0z\xb8[\x9e!\xf2h\x90\x8d\x82f%G4\xc2\xf0'
            b'\xda\x8dr\x81ER?S6\x12%7\x98\x89\x90'
        )
        expected_message = ('The encryption key must be of type "bytes", '
                            'not "<class \'str\'>"')

        with self.assertRaisesMessage(TypeError, expected_message):
            aes_decrypt(encrypted, key='abc')

    def test_aes_encrypt_with_bytes(self):
        """Testing aes_encrypt with byte string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt(self.PLAIN_TEXT_BYTES)
        self.assertIsInstance(encrypted, bytes)
        self.assertEqual(aes_decrypt(encrypted), self.PLAIN_TEXT_BYTES)

    def test_aes_encrypt_with_unicode(self):
        """Testing aes_encrypt with Unicode string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt(self.PLAIN_TEXT_UNICODE)
        self.assertIsInstance(encrypted, bytes)
        self.assertEqual(aes_decrypt(encrypted), self.PLAIN_TEXT_BYTES)

    def test_aes_encrypt_with_custom_key(self):
        """Testing aes_encrypt with custom key"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt(self.PLAIN_TEXT_BYTES, key=self.CUSTOM_KEY)
        decrypted = aes_decrypt(encrypted, key=self.CUSTOM_KEY)

        self.assertIsInstance(decrypted, bytes)
        self.assertEqual(decrypted, self.PLAIN_TEXT_BYTES)

    def test_decrypt_password_with_bytes(self):
        """Testing decrypt_password with byte string"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = b'AjsUGevO3UiVH7iN3zO9vxvqr5X5ozuAbOUByTATsitkhsih1Zc='
        decrypted = decrypt_password(encrypted)

        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)

    def test_decrypt_password_with_unicode(self):
        """Testing decrypt_password with Unicode string"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = 'AjsUGevO3UiVH7iN3zO9vxvqr5X5ozuAbOUByTATsitkhsih1Zc='
        decrypted = decrypt_password(encrypted)

        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)

    def test_decrypt_password_with_custom_key(self):
        """Testing decrypt_password with custom key"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = b'/pOO3VWHRXd1ZAeHZo8MBGQsNClD4lS7XK9WAydt8zW/ob+e63E='
        decrypted = decrypt_password(encrypted, key=self.CUSTOM_KEY)

        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)

    def test_encrypt_password_with_bytes(self):
        """Testing encrypt_password with byte string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = decrypt_password(encrypt_password(self.PASSWORD_BYTES))
        self.assertIsInstance(encrypted, str)
        self.assertEqual(encrypted, self.PASSWORD_UNICODE)

    def test_encrypt_password_with_unicode(self):
        """Testing encrypt_password with Unicode string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = decrypt_password(encrypt_password(self.PASSWORD_UNICODE))
        self.assertIsInstance(encrypted, str)
        self.assertEqual(encrypted, self.PASSWORD_UNICODE)

    def test_encrypt_password_with_custom_key(self):
        """Testing encrypt_password with custom key"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = encrypt_password(self.PASSWORD_UNICODE,
                                     key=self.CUSTOM_KEY)
        self.assertIsInstance(encrypted, str)

        decrypted = decrypt_password(encrypted, key=self.CUSTOM_KEY)
        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)

    def test_get_default_aes_encryption_key(self):
        """Testing get_default_aes_encryption_key"""
        key = get_default_aes_encryption_key()
        self.assertIsInstance(key, bytes)
        self.assertEqual(key, b'abcdefghijklmnop')



"""Unit tests for reviewboard.datagrids.columns.ReviewCountColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import ReviewCountColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class ReviewCountColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.ReviewCountColumn.

    Version Added:
        8.0
    """

    column = ReviewCountColumn()

    def test_render_data_with_zero_reviews(self) -> None:
        """Testing ReviewCountColumn.render_data with zero reviews"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        review_request.publicreviewcount_count = 0

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '0')

    def test_render_data_with_reviews(self) -> None:
        """Testing ReviewCountColumn.render_data with reviews"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        review_request.publicreviewcount_count = 3

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '3')

    def test_to_json_with_zero_reviews(self) -> None:
        """Testing ReviewCountColumn.to_json with zero reviews"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        review_request.publicreviewcount_count = 0

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            0)

    def test_to_json_with_reviews(self) -> None:
        """Testing ReviewCountColumn.to_json with reviews"""
        review_request = self.create_review_request(publish=True)

        # This is normally set by the column's augment_queryset().
        review_request.publicreviewcount_count = 3

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            3)



"""Views for reviewing file attachments (and legacy screenshots)."""

from __future__ import annotations

import logging
from typing import Optional, TYPE_CHECKING

from django.db.models import Q
from django.http import (Http404,
                         HttpRequest,
                         HttpResponse,
                         HttpResponseRedirect)
from django.shortcuts import get_object_or_404
from django.views.generic.base import View

from reviewboard.accounts.mixins import UserProfileRequiredViewMixin
from reviewboard.attachments.models import FileAttachment
from reviewboard.reviews.context import should_view_draft
from reviewboard.reviews.models import Screenshot
from reviewboard.reviews.ui.base import (DiffMismatchReviewUI,
                                         ReviewUI)
from reviewboard.reviews.ui.screenshot import LegacyScreenshotReviewUI
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin


if TYPE_CHECKING:
    from reviewboard.reviews.models import (ReviewRequest,
                                            ReviewRequestDraft)


logger = logging.getLogger(__name__)


class _FileAttachmentViewMixin:
    """Mixin for file attachment views.

    Version Added:
        7.0
    """

    def get_file_attachment(
        self,
        *,
        review_request: ReviewRequest,
        draft: Optional[ReviewRequestDraft],
        file_attachment_id: int,
        extra_q: Optional[Q] = None,
    ) -> FileAttachment:
        """Return a file attachment accessible on the review request.

        This will check for the file attachment on the review request or
        draft, both active and inactive file attachments.

        If the file attachment could not be found, this will raise a
        :http:`404`.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request the file attachment is on.

            draft (reviewboard.reviews.models.ReviewRequestDraft):
                The optional draft that the file attachment may (or may not)
                be on.

            file_attachment_id (int):
                The ID of the file attachment.

            extra_q (django.db.models.Q, optional):
                Extra query arguments to include for filtering.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The file attachment matching the criteria.

        Raises:
            django.http.Http404:
                The file attachment was not found.
        """
        # Make sure the attachment returned is part of either the review
        # request or an accessible draft.
        review_request_q = (Q(review_request=review_request) |
                            Q(inactive_review_request=review_request))

        if extra_q is None:
            extra_q = Q()

        if draft is not None:
            review_request_q |= Q(drafts=draft) | Q(inactive_drafts=draft)

        return get_object_or_404(
            FileAttachment,
            Q(pk=file_attachment_id) & extra_q & review_request_q)


class DownloadFileAttachmentView(_FileAttachmentViewMixin,
                                 ReviewRequestViewMixin,
                                 View):
    """Redirects the request to the file attachment on the storage backend.

    This will check first for an accessible file attachment matching the
    URL and the user's access to the review request and draft. If found, the
    client will be redirected to the location of the file attachment in the
    storage backend.

    The redirected file may or may not be cacheable, and may only be
    accessible temporarily, depending on the backend.

    Version Added:
        7.0
    """

    def get(
        self,
        request: HttpRequest,
        *,
        file_attachment_id: int,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            file_attachment_id (int):
                The revision of the file attachment to download the file from.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response to send to the client.
        """
        user = request.user
        review_request = self.review_request

        if user is not None:
            draft = review_request.get_draft(user)
        else:
            draft = None

        file_attachment = self.get_file_attachment(
            review_request=review_request,
            draft=draft,
            file_attachment_id=file_attachment_id)

        download_url: Optional[str] = None

        if request.GET.get('thumbnail'):
            # The caller requested a thumbnail. Determine the requested size.
            try:
                thumbnail_width = int(request.GET.get('width', ''))
            except ValueError:
                thumbnail_width = None

            try:
                thumbnail_height = int(request.GET.get('height', ''))
            except ValueError:
                thumbnail_height = None

            if thumbnail_width or thumbnail_height:
                download_url = file_attachment.get_raw_thumbnail_image_url(
                    width=thumbnail_width,
                    height=thumbnail_height)
        else:
            download_url = file_attachment.get_raw_download_url()

        if download_url:
            return HttpResponseRedirect(download_url)

        raise Http404


class ReviewFileAttachmentView(_FileAttachmentViewMixin,
                               ReviewRequestViewMixin,
                               UserProfileRequiredViewMixin,
                               View):
    """Displays a file attachment with a review UI."""

    def get(
        self,
        request: HttpRequest,
        file_attachment_id: int,
        file_attachment_diff_id: Optional[int] = None,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle a HTTP GET request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            file_attachment_id (int):
                The ID of the file attachment to review.

            file_attachment_diff_id (int, optional):
                The ID of the file attachment to diff against.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response from the handler.
        """
        review_request = self.review_request
        user = request.user
        draft = review_request.get_draft(user)

        if should_view_draft(request=request, review_request=review_request,
                             draft=draft):
            file_attachment = self.get_file_attachment(
                review_request=review_request,
                draft=draft,
                file_attachment_id=file_attachment_id)
        else:
            file_attachment = self.get_file_attachment(
                review_request=review_request,
                draft=None,
                file_attachment_id=file_attachment_id)

        diff_against_attachment: Optional[FileAttachment] = None

        review_ui_class = ReviewUI.for_object(file_attachment)

        if file_attachment_diff_id:
            diff_against_attachment = self.get_file_attachment(
                review_request=review_request,
                draft=draft,
                file_attachment_id=file_attachment_diff_id,
                extra_q=Q(
                    attachment_history=file_attachment.attachment_history
                ))

            diff_review_ui_class = ReviewUI.for_object(diff_against_attachment)

            if review_ui_class is not diff_review_ui_class:
                # The file types between the original and modified attachments
                # don't match. Just use the base ReviewUI to render, which will
                # show an error message.
                dummy_review_ui = DiffMismatchReviewUI(
                    review_request=review_request,
                    obj=file_attachment)
                dummy_review_ui.set_diff_against(diff_against_attachment)

                return dummy_review_ui.render_to_response(request)

        if (review_ui_class is None or
            (diff_against_attachment and
             not review_ui_class.supports_diffing)):
            raise Http404

        review_ui = review_ui_class(
            review_request=review_request,
            obj=file_attachment)

        if diff_against_attachment:
            review_ui.set_diff_against(diff_against_attachment)

        if not file_attachment.is_review_ui_accessible_by(user):
            raise Http404

        if (diff_against_attachment and
            not diff_against_attachment.is_review_ui_accessible_by(user)):
            raise Http404

        return review_ui.render_to_response(request)


class ReviewScreenshotView(ReviewRequestViewMixin,
                           UserProfileRequiredViewMixin,
                           View):
    """Displays a review UI for a screenshot.

    Screenshots are a legacy feature, predating file attachments. While they
    can't be created anymore, this view does allow for reviewing screenshots
    uploaded in old versions.
    """

    def get(
        self,
        request: HttpRequest,
        screenshot_id: int,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle a HTTP GET request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            screenshot_id (int):
                The ID of the screenshot to review.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response from the handler.
        """
        review_request = self.review_request
        draft = review_request.get_draft(request.user)

        # Make sure the screenshot returned is part of either the review
        # request or an accessible draft.
        review_request_q = (Q(review_request=review_request) |
                            Q(inactive_review_request=review_request))

        if draft:
            review_request_q |= Q(drafts=draft) | Q(inactive_drafts=draft)

        screenshot = get_object_or_404(Screenshot,
                                       Q(pk=screenshot_id) & review_request_q)
        review_ui = LegacyScreenshotReviewUI(review_request, screenshot)

        return review_ui.render_to_response(request)



"""Unit tests for reviewboard.accounts.managers.ReviewRequestVisitManager."""

from __future__ import annotations

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.db.models import Q

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.testing import TestCase


class ReviewRequestVisitTests(TestCase):
    """Unit tests for reviewboard.accounts.managers.ReviewRequestVisitManager.
    """

    fixtures = ['test_users']

    def test_update_visibility_create(self):
        """Testing ReviewRequestVisitManager.update_visibility
        creates a new visit
        """
        review_request = self.create_review_request(publish=True)
        user = User.objects.get(username='admin')

        visit = ReviewRequestVisit.objects.update_visibility(
            review_request, user, ReviewRequestVisit.ARCHIVED)

        self.assertEqual(visit.visibility, ReviewRequestVisit.ARCHIVED)

    def test_update_visibility_update_visible(self):
        """Testing ReviewRequestVisitManager.update_visibility
        updates existing visit with visible
        """
        review_request = self.create_review_request(publish=True)
        user = User.objects.get(username='admin')

        visit = ReviewRequestVisit.objects.create(
            review_request=review_request, user=user,
            visibility=ReviewRequestVisit.VISIBLE)

        queries = [
            {
                'model': ReviewRequestVisit,
                'select_for_update': True,
                'where': Q(review_request=review_request, user=user),
            },
            {
                'model': ReviewRequestVisit,
                'type': 'UPDATE',
                'where': Q(pk=visit.pk),
            },
        ]

        with assert_queries(queries, num_statements=4):
            visit = ReviewRequestVisit.objects.update_visibility(
                review_request, user, ReviewRequestVisit.VISIBLE)

        self.assertEqual(visit.visibility, ReviewRequestVisit.VISIBLE)

    def test_update_visibility_update_archive(self):
        """Testing ReviewRequestVisitManager.update_visibility
        updates existing visit with archive
        """
        review_request = self.create_review_request(publish=True)
        user = User.objects.get(username='admin')

        ReviewRequestVisit.objects.create(
            review_request=review_request, user=user,
            visibility=ReviewRequestVisit.VISIBLE)
        visit = ReviewRequestVisit.objects.update_visibility(
            review_request, user, ReviewRequestVisit.ARCHIVED)

        self.assertEqual(visit.visibility, ReviewRequestVisit.ARCHIVED)



"""Unit tests for reviewboard.attachments.mimetypes.ImageMimetype.

Version Added:
    7.0.3:
    This was split off from :py:mod:`reviewboard.attachments.tests`.
"""

from __future__ import annotations

import os

import kgb

from reviewboard.attachments.forms import UploadFileForm
from reviewboard.attachments.mimetypes import logger as mimetypes_logger
from reviewboard.attachments.tests.base import BaseFileAttachmentTestCase


class ImageMimetypeTests(kgb.SpyAgency, BaseFileAttachmentTestCase):
    """Unit tests for reviewboard.attachments.mimetypes.ImageMimetype.

    Version Added:
        6.0
    """

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        image_file = self.make_uploaded_file()

        review_request = self.create_review_request(publish=True)

        form = UploadFileForm(review_request, files={
            'path': image_file,
        })
        self.assertTrue(form.is_valid())

        self.file_attachment = form.create()

    def test_get_thumbnail(self) -> None:
        """Testing ImageMimetype.get_thumbnail"""
        file = self.file_attachment.file
        storage = file.storage
        filename_base = os.path.splitext(file.name)[0]
        download_url = 'http://example.com/r/1/file/1/download/?thumbnail=1'

        self.assertHTMLEqual(
            self.file_attachment.thumbnail,
            f'<div class="file-thumbnail">'
            f'<img src="{download_url}&width=300"'
            f' srcset="{download_url}&width=300 1x,'
            f' {download_url}&width=600 2x, {download_url}&width=900 3x"'
            f'alt="" width="300" />'
            f'</div>')

        # These shouldn't exist until the URLs are accessed.
        self.assertFalse(storage.exists(f'{filename_base}_300.png'))
        self.assertFalse(storage.exists(f'{filename_base}_600.png'))
        self.assertFalse(storage.exists(f'{filename_base}_900.png'))

    def test_get_raw_thumbnail_image_url(self) -> None:
        """Testing ImageMimetype.get_raw_thumbnail_image_url"""
        file = self.file_attachment.file
        storage = file.storage
        filename_base = os.path.splitext(file.name)[0]

        mimetype_handler = self.file_attachment.mimetype_handler
        assert mimetype_handler

        filename = mimetype_handler.get_raw_thumbnail_image_url(width=300)
        self.assertTrue(filename)
        self.assertTrue(storage.exists(f'{filename_base}_300.png'))

    def test_generate_thumbnail_image_with_create_if_missing_false(
        self,
    ) -> None:
        """Testing ImageMimetype.get_raw_thumbnail_image_url with
        create_if_missing=False
        """
        file = self.file_attachment.file
        storage = file.storage
        filename_base = os.path.splitext(file.name)[0]

        mimetype_handler = self.file_attachment.mimetype_handler
        assert mimetype_handler

        filename = mimetype_handler.generate_thumbnail_image(
            width=300,
            create_if_missing=False)

        self.assertIsNone(filename)
        self.assertFalse(storage.exists(f'{filename_base}_300.png'))

    def test_delete_associated_files(self) -> None:
        """Testing ImageMimetype.delete_associated_files"""
        file = self.file_attachment.file
        storage = file.storage

        mimetype_handler = self.file_attachment.mimetype_handler
        assert mimetype_handler

        filename_base = os.path.splitext(file.name)[0]
        filename_300 = f'{filename_base}_300.png'
        filename_600 = f'{filename_base}_600.png'

        mimetype_handler.generate_thumbnail_image(width=300)
        mimetype_handler.generate_thumbnail_image(width=600)

        self.assertTrue(storage.exists(filename_300))
        self.assertTrue(storage.exists(filename_600))

        mimetype_handler.delete_associated_files()

        self.assertFalse(storage.exists(filename_300))
        self.assertFalse(storage.exists(filename_600))

        self.spy_on(storage.delete)
        self.spy_on(mimetypes_logger.warning)

        # A second call shouldn't do anything, outside of checking thumbnail
        # existence.
        mimetype_handler.delete_associated_files()

        self.assertSpyNotCalled(storage.delete)



"""Unit tests for reviewboard.extensions.hooks.WebAPICapabilitiesHook."""

from reviewboard.extensions.base import Extension
from reviewboard.extensions.hooks import WebAPICapabilitiesHook
from reviewboard.extensions.tests.testcases import ExtensionHookTestCaseMixin
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import root_item_mimetype
from reviewboard.webapi.tests.urls import get_root_url


class WebAPICapabilitiesExtension(Extension):
    metadata = {
        'Name': 'Web API Capabilities Extension',
    }


class WebAPICapabilitiesHookTests(ExtensionHookTestCaseMixin,
                                  BaseWebAPITestCase):
    """Testing WebAPICapabilitiesHook."""

    extension_class = WebAPICapabilitiesExtension

    def setUp(self):
        super(WebAPICapabilitiesHookTests, self).setUp()

        self.old_extension_id = self.extension.id
        self.url = get_root_url()

    def tearDown(self):
        self.extension.id = self.old_extension_id

        super(WebAPICapabilitiesHookTests, self).tearDown()

    def test_register(self):
        """Testing WebAPICapabilitiesHook initializing"""
        WebAPICapabilitiesHook(
            extension=self.extension,
            caps={
                'sandboxed': True,
                'thorough': True,
            })

        rsp = self.api_get(path=self.url,
                           expected_mimetype=root_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('capabilities', rsp)

        caps = rsp['capabilities']
        self.assertIn(self.extension.id, caps)

        extension_caps = caps[self.extension.id]
        self.assertTrue(extension_caps['sandboxed'])
        self.assertTrue(extension_caps['thorough'])

    def test_register_fails_no_id(self):
        """Testing WebAPICapabilitiesHook initializing with ID of None"""
        self.extension.id = None

        expected_message = \
            'The capabilities_id attribute must not be None'

        with self.assertRaisesMessage(ValueError, expected_message):
            WebAPICapabilitiesHook(self.extension, {
                'sandboxed': True,
                'thorough': True,
            })

        rsp = self.api_get(path=self.url,
                           expected_mimetype=root_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('capabilities', rsp)

        caps = rsp['capabilities']
        self.assertNotIn(self.extension.id, caps)
        self.assertNotIn(None, caps)

    def test_register_fails_default_capability(self):
        """Testing WebAPICapabilitiesHook initializing with default key"""
        self.extension.id = 'diffs'

        expected_message = \
            '"diffs" is reserved for the default set of capabilities'

        with self.assertRaisesMessage(KeyError, expected_message):
            WebAPICapabilitiesHook(self.extension, {
                'base_commit_ids': False,
                'moved_files': False,
            })

        rsp = self.api_get(path=self.url,
                           expected_mimetype=root_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('capabilities', rsp)

        caps = rsp['capabilities']
        self.assertIn('diffs', caps)

        diffs_caps = caps['diffs']
        self.assertTrue(diffs_caps['base_commit_ids'])
        self.assertTrue(diffs_caps['moved_files'])

    def test_unregister(self):
        """Testing WebAPICapabilitiesHook uninitializing"""
        hook = WebAPICapabilitiesHook(
            extension=self.extension,
            caps={
                'sandboxed': True,
                'thorough': True,
            })

        hook.disable_hook()

        rsp = self.api_get(path=self.url,
                           expected_mimetype=root_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('capabilities', rsp)

        caps = rsp['capabilities']
        self.assertNotIn(self.extension.id, caps)



#!/usr/bin/env python

from __future__ import annotations

import os
import shutil
import subprocess
import sys
from pathlib import Path

import django
from django.core.management import CommandError, call_command
from pipeline.exceptions import CompilerError


scripts_dir = os.path.abspath(os.path.dirname(__file__))

# Source root directory
root_dir = os.path.abspath(os.path.join(scripts_dir, '..', '..'))
sys.path.insert(0, root_dir)

# Script config directory
sys.path.insert(0, os.path.join(scripts_dir, 'conf'))


if __name__ == '__main__':
    os.chdir(root_dir)

    # Verify that we have npm.
    npm_command = 'npm'

    try:
        subprocess.check_call([npm_command, '--version'],
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        raise RuntimeError(
            f'Unable to locate {npm_command} in the path, which is needed to '
            f'compile static media.'
        )

    # Install dependencies.
    subprocess.call([npm_command, 'install'])

    # Set up the Django environment.
    os.environ['FORCE_BUILD_MEDIA'] = '1'
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'reviewboard.settings')

    django.setup()

    # Check if we're actually building media. This internal flag is used to
    # by the package build backend to better control setup vs. building of
    # static media.
    if os.environ.get('RUN_COLLECT_STATIC') != '0':
        # Remove any stale htdocs files.
        htdocs_static_dir = os.path.join(root_dir, 'reviewboard', 'htdocs',
                                         'static')

        if os.path.exists(htdocs_static_dir):
            shutil.rmtree(htdocs_static_dir)

        from django.conf import settings

        jquery_ui_dir: (Path | None) = None

        for node_path in settings.NODE_PATH.split(':'):
            jquery_ui = Path(node_path) / 'jquery-ui'

            if jquery_ui.is_dir():
                jquery_ui_dir = jquery_ui

        # We currently bundle jquery-ui.css out of the base theme. This
        # stylesheet references a few images, which need to be collected.
        #
        # This can go away once we get rid of our jquery-ui dependency.
        assert jquery_ui_dir is not None
        jquery_ui_images = \
            jquery_ui_dir / 'dist' / 'themes' / 'base' / 'images'

        settings.STATICFILES_FINDERS.append(
            'django.contrib.staticfiles.finders.FileSystemFinder',
        )

        # We completely override STATICFILES_DIRS in this case because we don't
        # want warnings about duplicate files between the regular finders and
        # the FileSystemFinder.
        settings.STATICFILES_DIRS = [
            ('lib/css/images', str(jquery_ui_images)),
        ]

        # Build the static media.
        #
        # This will raise a CommandError or call sys.exit(1) on failure.
        try:
            call_command('collectstatic', interactive=False, verbosity=2)
        except CompilerError as e:
            sys.stderr.write(e.error_output)
            sys.stderr.write('\n')

            raise CommandError('Failed to build static media')



"""OAuth2-related URLs."""

from django.urls import path
from oauth2_provider import views as oauth2_provider_views

from reviewboard.oauth import views


urlpatterns = [
    path('authorize/', views.AuthorizationView.as_view(), name='authorize'),
    path('token/', oauth2_provider_views.TokenView.as_view(), name='token'),
]



from django import forms
from django.utils.translation import gettext_lazy as _

from reviewboard.admin.validation import validate_bug_tracker_base_hosting_url
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService


class TracForm(BaseHostingServiceRepositoryForm):
    trac_url = forms.CharField(
        label=_('Trac URL'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        validators=[validate_bug_tracker_base_hosting_url])

    def clean_trac_url(self):
        return self.cleaned_data['trac_url'].rstrip('/')


class Trac(BaseHostingService):
    name = 'Trac'
    hosting_service_id = 'trac'
    form = TracForm
    bug_tracker_field = '%(trac_url)s/ticket/%%s'
    supports_bug_trackers = True



"""Extension hooks for customizing email notifications."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

from reviewboard.notifications.email import (register_email_hook,
                                             unregister_email_hook)
from reviewboard.reviews.signals import (review_request_published,
                                         review_published, reply_published,
                                         review_request_closed)


class EmailHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for changing the recipients of e-mails.

    Extensions can use this hook to change the contents of the To and CC fields
    of e-mails. This should be subclassed in an extension to provide the
    desired behaviour. This class is a base class for more specialized
    extension hooks. If modifying only one type of e-mail's fields is desired,
    one of the following classes should be subclassed instead.

    * :py:class:`ReviewPublishedEmailHook`
    * :py:class:`ReviewReplyPublishedEmailHook`
    * :py:class:`ReviewRequestPublishedEmailHook`
    * :py:class:`ReviewRequestClosedEmailHook`

    However, if more specialized behaviour is desired, this class can be
    subclassed.
    """

    def initialize(self, signals):
        """Initialize the hook.

        Args:
            signals (list):
                A list of :py:class:`Signals <django.dispatch.Signal>` that,
                when triggered, will cause e-mails to be sent. Valid signals
                are:

                * :py:data:`~reviewboard.reviews.signals.review_request_published`
                * :py:data:`~reviewboard.reviews.signals.review_request_closed`
                * :py:data:`~reviewboard.reviews.signals.review_published`
                * :py:data:`~reviewboard.reviews.signals.reply_published`
        """
        self.signals = signals

        for signal in signals:
            register_email_hook(signal, self)

    def shutdown(self):
        """Shut down the hook.

        This will unregister each of the e-mail handlers.
        """
        for signal in self.signals:
            unregister_email_hook(signal, self)

    def get_to_field(self, to_field, **kwargs):
        """Return the To field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>`
                that will receive the e-mail.

            kwargs (dict):
                Additional keyword arguments that will be passed based on the
                type of e-mail being sent.

        Returns:
            set: The desired To field.
        """
        return to_field

    def get_cc_field(self, cc_field, **kwargs):
        """Return the CC field for the e-mail.

        Args:
            cc_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>`
                that will receive a carbon copy of the e-mail.

            kwargs (dict):
                Additional keyword arguments that will be passed based on the
                type of e-mail being sent.

        Returns:
            set: The desired CC field.
        """
        return cc_field


class ReviewPublishedEmailHook(EmailHook):
    """A hook for changing the recipients of review publishing e-mails.

    This hook must be subclassed. The caller is expected to override
    :py:meth:`get_to_field` and/or :py:meth:`get_cc_field`.
    """

    def initialize(self):
        """Initialize the hook."""
        super(ReviewPublishedEmailHook, self).initialize(
            signals=[review_published])

    def get_to_field(self, to_field, review, user, review_request,
                     to_owner_only, **kwargs):
        """Return the To field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>`
                that will receive the e-mail.

            review (reviewboard.reviews.models.Review):
                The review that was published.

            user (django.contrib.auth.models.User):
                The user who published the review.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was reviewed.

            to_owner_only (bool):
                Whether or not the review was marked as being targeted at only
                the submitter.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired To field.
        """
        return to_field

    def get_cc_field(self, cc_field, review, user, review_request,
                     to_owner_only, **kwargs):
        """Return the CC field for the e-mail.

        Args:
            cc_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>`
                that will receive a carbon copy of the e-mail.

            review (reviewboard.reviews.models.Review):
                The review that was published.

            user (django.contrib.auth.models.User):
                The user who published the review.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was reviewed.

            to_owner_only (bool):
                Whether or not the review was marked as being targeted at only
                the submitter.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired CC field.
        """
        return cc_field


class ReviewReplyPublishedEmailHook(EmailHook):
    """A hook for changing the recipients of review reply publishing e-mails.

    This hook must be subclassed. The caller is expected to override
    :py:meth:`get_to_field` and/or :py:meth:`get_cc_field`.
    """

    def initialize(self):
        """Initialize the hook."""
        super(ReviewReplyPublishedEmailHook, self).initialize(
            signals=[reply_published])

    def get_to_field(self, to_field, reply, user, review_request, **kwargs):
        """Return the To field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>`
                that will receive the e-mail.

            reply (reviewboard.reviews.models.Review):
                The review reply that was published.

            user (django.contrib.auth.models.User):
                The user who published the review reply.

            review (reviewboard.reviews.model.Review):
                The review the reply is in reply to.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was reviewed.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired To field.
        """
        return to_field

    def get_cc_field(self, cc_field, reply, user, review_request, **kwargs):
        """Return the CC field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>`
                that will receive a carbon copy of the e-mail

            reply (reviewboard.reviews.models.Review):
                The review reply that was published.

            user (django.contrib.auth.models.User):
                The user who published the reply.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was reviewed.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired CC field.
        """
        return cc_field


class ReviewRequestClosedEmailHook(EmailHook):
    """A hook for changing the recipients of review request closing e-mails.

    This hook must be subclassed. The caller is expected to override
    :py:meth:`get_to_field` and/or :py:meth:`get_cc_field`.
    """

    def initialize(self):
        """Initialize the hook."""
        super(ReviewRequestClosedEmailHook, self).initialize(
            signals=[review_request_closed])

    def get_to_field(self, to_field, review_request, user, close_type,
                     **kwargs):
        """Return the To field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>`
                that will receive the e-mail.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was published.

            user (django.contrib.auth.models.User):
                The user who closed the review request.

            close_type (unicode):
                How the review request was closed. This is one of
                :py:attr:`~reviewboard.reviews.models.ReviewRequest.SUBMITTED`
                or
                :py:attr:`~reviewboard.reviews.models.ReviewRequest.DISCARDED`.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired To field.
        """
        return to_field

    def get_cc_field(self, cc_field, review_request, user, close_type,
                     **kwargs):
        """Return the CC field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>` that
                will receive a carbon copy of the e-mail.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was published.

            user (django.contrib.auth.models.User):
                The user who closed the review request.

            close_type (unicode):
                How the review request was closed. This is one of
                :py:attr:`~reviewboard.reviews.models.ReviewRequest.SUBMITTED`
                or
                :py:attr:`~reviewboard.reviews.models.ReviewRequest.DISCARDED`.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired CC field.
        """
        return cc_field


class ReviewRequestPublishedEmailHook(EmailHook):
    """A hook for changing the recipients of review request publishing e-mails.

    This hook must be subclassed. The caller is expected to override
    :py:meth:`get_to_field` and/or :py:meth:`get_cc_field`.
    """

    def initialize(self):
        """Initialize the hook. """
        super(ReviewRequestPublishedEmailHook, self).initialize(
            signals=[review_request_published])

    def get_to_field(self, to_field, review_request, user, **kwargs):
        """Return the To field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>` that
                will receive the e-mail.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was published.

            user (django.contrib.auth.models.User):
                The user who published the review request.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired To field.
        """
        return to_field

    def get_cc_field(self, cc_field, review_request, user, **kwargs):
        """Return the CC field for the e-mail.

        Args:
            to_field (set):
                A set of :py:class:`Users <django.contrib.auth.models.User>`
                and :py:class:`Groups <reviewboard.reviews.models.Group>` that
                will receive a carbon copy of the e-mail.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that was published.

            user (django.contrib.auth.models.User):
                The user who published the review request.

            **kwargs (dict):
                Additional keyword arguments, since the signature may change in
                the future.

        Returns:
            set: The desired CC field.
        """
        return cc_field



"""Certificates, fingerprints, and bundles.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
import os
import re
import socket
import ssl

from datetime import datetime, timezone
from enum import Enum
from ipaddress import (IPv4Address,
                       IPv4Network,
                       IPv6Address,
                       IPv6Network,
                       ip_address)
from typing import TYPE_CHECKING, Union
from uuid import uuid4

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.x509.oid import NameOID
from django.utils import timezone as django_timezone
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.translation import gettext as _
from djblets.util.symbols import UNSET, Unsettable
from typelets.runtime import raise_invalid_type
from typing_extensions import Self

from reviewboard.certs.errors import (CertificateNotFoundError,
                                      CertificateStorageError,
                                      InvalidCertificateFormatError)
from reviewboard.certs.utils import (get_cert_hostname_matches,
                                     normalize_cert_hostname)

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Final, TypeAlias, TypeVar

    from typelets.django.json import (SerializableDjangoJSONDict,
                                      SerializableDjangoJSONDictImmutable)

    _T = TypeVar('_T')


logger = logging.getLogger(__name__)


_CERT_PEM_RE = re.compile(
    br'-----BEGIN CERTIFICATE-----[\r\n]+'
    br'[A-Za-z0-9+/\r\n]+=*[\r\n]+'
    br'-----END CERTIFICATE-----'
)

_PRIVATE_KEY_PEM_RE = re.compile(
    br'-----BEGIN PRIVATE KEY-----[\r\n]+'
    br'[A-Za-z0-9+/\r\n]+=*[\r\n]+'
    br'-----END PRIVATE KEY-----'
)


#: A value in a certificate SAN field.
#:
#: Version Added:
#:     8.0
SANValue: TypeAlias = Union[
    IPv4Address,
    IPv4Network,
    IPv6Address,
    IPv6Network,
    str,
]


def _format_fingerprint(
    fingerprint: bytes,
) -> str:
    """Return a string representation of a fingerprint.

    This will be in ``AA:BB:CC...`` format.

    Version Added:
        6.0

    Args:
        fingerprint (bytes):
            The raw fingerprint data to format.

    Returns:
        str:
        The formatted string representation.
    """
    return ':'.join(
        '%02X' % c
        for c in fingerprint
    )


class CertDataFormat(Enum):
    """Certificate data formats.

    Version Added:
        6.0
    """

    #: PEM-formatted certificate data.
    PEM = 'PEM'


class CertPurpose(Enum):
    """The purpose of a certificate.

    Version Added:
        8.0
    """

    #: A certificate used to verify and trust a server.
    TRUST = 'trust'

    #: A certificate used to identify the client for authentication.
    CLIENT = 'client'


class CertificateFingerprints:
    """Representation of certificate fingerprints.

    Version Added:
        6.0
    """

    #: Regex for matching a SHA1 fingerprint.
    #:
    #: Version Added:
    #:     8.0
    SHA1_FINGERPRINT_RE = re.compile(r'(?:[0-9A-F]{2}:){19}[0-9A-F]{2}',
                                     re.IGNORECASE)

    #: Regex for matching a SHA1 string.
    #:
    #: Version Added:
    #:     8.0
    SHA1_RE = re.compile(r'[0-9A-F]{40}',
                         re.IGNORECASE)

    #: Regex for matching a SHA256 fingerprint.
    #:
    #: Version Added:
    #:     8.0
    SHA256_FINGERPRINT_RE = re.compile(r'(?:[0-9A-F]{2}:){31}[0-9A-F]{2}',
                                       re.IGNORECASE)

    #: Regex for matching a SHA256 string.
    #:
    #: Version Added:
    #:     8.0
    SHA256_RE = re.compile(r'[0-9A-F]{64}',
                           re.IGNORECASE)

    #: Mapping of string lengths to pattern/attr/normalization flags.
    #:
    #: This works as a simple table to quickly locate a proper pattern,
    #: corresponding attribute for the constructor, and a flag indicating
    #: whether to perform normalization of the value.
    #:
    #: Version Added:
    #:     8.0
    _FINGERPRINT_INFO_BY_LENGTH: Final[Mapping[
        int,
        tuple[re.Pattern, str, bool],
    ]] = {
        40: (SHA1_RE, 'sha1', True),
        59: (SHA1_FINGERPRINT_RE, 'sha1', False),
        64: (SHA256_RE, 'sha256', True),
        95: (SHA256_FINGERPRINT_RE, 'sha256', False),
    }

    ######################
    # Instance variables #
    ######################

    #: The human-readable SHA1 fingerprint.
    sha1: (str | None) = None

    #: The human-readable SHA256 fingerprint.
    sha256: (str | None) = None

    @classmethod
    def from_json(
        cls,
        data: SerializableDjangoJSONDictImmutable,
    ) -> Self:
        """Return a new instance from a serialized JSON payload.

        The payload is expected to be in the following format:

        Keys:
            sha1 (str, optional):
                The human-readable SHA1 fingerprint in ``AA:BB:CC...`` form.

            sha256 (str, optional):
                The human-readable SHA256 fingerprint in ``AA:BB:CC...`` form.

        Args:
            data (dict):
                The JSON dictionary containing the fingerprint information.

        Returns:
            CertificateFingerprints:
            The parsed fingerprints instance.
        """
        sha1 = data.get('sha1') or None
        sha256 = data.get('sha256') or None

        if sha1 is not None and not isinstance(sha1, str):
            logger.warning('Got non-string value %r for "sha1" key for '
                           'fingerprint data=%r',
                           sha1, data)
            sha1 = None

        if sha256 is not None and not isinstance(sha256, str):
            logger.warning('Got non-string value %r for "sha256" key for '
                           'fingerprint data=%r',
                           sha256, data)
            sha256 = None

        return cls(sha1=sha1,
                   sha256=sha256)

    @classmethod
    def from_string(
        cls,
        fingerprint: str,
    ) -> Self | None:
        """Return a new instance from a single fingerprint string.

        This will normalize the provided fingerprint (stripping any
        whitespace), determine if this is SHA-1 or SHA-256, and then return a
        new instance if the value is supported.

        Version Added:
            8.0

        Args:
            fingerprint (str):
                The fingerprint string in ``AA:BB:CC...`` format.

        Returns:
            reviewboard.certs.cert.CertificateFingerprints:
            The resulting fingerprints instance, or ``None`` if the
            string length doesn't match a known algorithm.
        """
        fingerprint = fingerprint.strip()
        length = len(fingerprint)

        # This will look for the normalized fingerprint format and the
        # raw SHA format for both SHA256 and SHA1 variations. To keep
        # this performant, we'll key off the fingerprint pattern, field,
        # and normalization flag by length. If the length is a match but
        # nothing else is, we don't want to proceed with other checks.
        try:
            sha_re, field, normalize = cls._FINGERPRINT_INFO_BY_LENGTH[length]

            if sha_re.fullmatch(fingerprint):
                if normalize:
                    fingerprint = ':'.join(
                        fingerprint[i:i + 2]
                        for i in range(0, length, 2)
                    )

                return cls(**{
                    field: fingerprint.upper()
                })
        except KeyError:
            pass

        return None

    @classmethod
    def from_x509_cert(
        cls,
        x509_cert: x509.Certificate,
    ) -> Self:
        """Return a new instance from a Cryptography X509 certificate.

        Args:
            x509_cert (cryptography.x509.Certificate):
                The Cryptography certificate used to load the fingerprints.

        Returns:
            CertificateFingerprints:
            The loaded fingerprints instance.
        """
        return cls(
            sha1=_format_fingerprint(x509_cert.fingerprint(hashes.SHA1())),
            sha256=_format_fingerprint(x509_cert.fingerprint(hashes.SHA256())))

    def __init__(
        self,
        *,
        sha1: (str | None) = None,
        sha256: (str | None) = None,
    ) -> None:
        """Initialize the certificate fingerprints instance.

        Args:
            sha1 (str):
                The SHA1 fingerprint in ``AA:BB:CC...`` format.

            shaw256 (str):
                The SHA256 fingerprint in ``AA:BB:CC...`` format.
        """
        self.sha1 = sha1
        self.sha256 = sha256

    def to_json(self) -> SerializableDjangoJSONDictImmutable:
        """Serialize the fingerprints to a JSON payload.

        Returns:
            dict:
            The resulting JSON payload, containing:

            Keys:
                sha1 (str, optional):
                    A human-readable SHA1 fingerprint in ``AA:BB:CC...``
                    form.

                sha256 (str, optional):
                    A human-readable SHA256 fingerprint in ``AA:BB:CC...``
                    form.

            These keys will only be present if there are fingerprints
            available.
        """
        data: SerializableDjangoJSONDict = {}

        if self.sha1:
            data['sha1'] = self.sha1

        if self.sha256:
            data['sha256'] = self.sha256

        return data

    def is_empty(self) -> bool:
        """Return whether these fingerprints are empty.

        Returns:
            bool:
            ``True`` if the fingerprints are empty (no fingerprints are
            stored). ``False`` if there are fingerprints available.
        """
        return not bool(self.sha1) and not bool(self.sha256)

    def matches(
        self,
        other: CertificateFingerprints,
    ) -> bool:
        """Return whether one set of fingerprints matches another.

        This will compare any available fingerprints between two instances,
        returning whether there's a match.

        Args:
            other (CertificateFingerprints):
                The other instance to compare to.

        Returns:
            bool:
            ``True`` if there is a match between two instances. ``False``
            if there is not.
        """
        if self.is_empty() or other.is_empty():
            return False

        has_sha1 = self.sha1 is not None and other.sha1 is not None
        has_sha256 = self.sha256 is not None and other.sha256 is not None

        if not has_sha1 and not has_sha256:
            # There's nothing we can compare.
            return False

        # We now know that there's something to compare. We'll either be
        # comparing Nones safely, or we'll be comparing actual fingerprints.
        sha1_match = not has_sha1 or self.sha1 == other.sha1
        sha256_match = not has_sha256 or self.sha256 == other.sha256

        return sha1_match and sha256_match

    def __eq__(
        self,
        other: object,
    ) -> bool:
        """Return whether this object is equal to another.

        Two objects are equal if they're both
        :py:class:`CertificateFingerprints` instances and contain the same
        signatures.

        Args:
            other (object):
                The object to compare this to.

        Returns:
            bool:
            ``True`` if they are equal. ``False`` if they are not.
        """
        return (
            isinstance(other, CertificateFingerprints) and
            self.sha1 == other.sha1 and
            self.sha256 == other.sha256
        )

    def __repr__(self) -> str:
        """Return a string representation of the instance.

        Returns:
            str:
            The string representation.
        """
        return (
            '<CertificateFingerprints(sha1=%(sha1)r, sha256=%(sha256)r)>'
            % {
                'sha1': self.sha1,
                'sha256': self.sha256,
            }
        )


class Certificate:
    """A representation of a SSL/TLS certificate.

    This may be an incomplete representation, with only the hostname and at
    least one fingerprint being required. It can be used to convey information
    about certificates from a server or tool, or used to provide data for
    storage.

    There are two types of certificates:

    * Trust certificates.

      These are used to verify a remote server's certificate during TLS.
      These may be root CA certificates, intermediate CA certificates,
      self-signed server certificates, or other certificates used as trust
      anchors.

      They do not include a private key.

    * Client certificates.

      These are used to authenticate Review Board with a remote service during
      mutual TLS. This requires an associated private key.

    Consumers should take care not to modify any certificate data after
    loading. While it's possible to change the data, doing so can lead to
    incorrect results, as some data is computed and then cached on the
    instance and cannot be updated later.

    Version Changed:
        8.0:
        Added explicit support for specifying a certificate purpose.

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The loaded certificate data.
    #:
    #: This will always be available for stored certificates, but may not be
    #: available as part of error responses.
    #:
    #: If available, it will match the format specified in
    #: :py:attr:`data_format`.
    cert_data: bytes | None

    #: The format for the loaded certificate and private key data.
    data_format: CertDataFormat

    #: The hostname that would serve this certificate.
    #:
    #: Note that this may be a wildcard domain (e.g., ``*.example.com``).
    hostname: str

    #: The loaded private key data, if available.
    #:
    #: This will match the format specified in :py:attr:`data_format`.
    key_data: bytes | None

    #: The port on the host that would serve this certificate.
    port: int

    #: The purpose set for a certificate.
    #:
    #: This defines whether the certificate is used for trusting a remote
    #: server or authenticating Review Board with a service.
    #:
    #: Version Added:
    #:     8.0
    purpose: CertPurpose

    #: Fingerprints for the certificate.
    #:
    #: If not provided during construction, this will be loaded from
    #: :py:attr`cert_data` when needed (and if :py:attr:`cert_data` is
    #: provided).
    _fingerprints: Unsettable[CertificateFingerprints | None]

    #: The issuer (usually the hostname) of the certificate.
    #:
    #: If not provided during construction, this will be loaded from
    #: :py:attr`cert_data` when needed (and if :py:attr:`cert_data` is
    #: provided).
    _issuer: Unsettable[str | None]

    #: The subject (usually the hostname) of the certificate.
    #:
    #: If not provided during construction, this will be loaded from
    #: :py:attr`cert_data` when needed (and if :py:attr:`cert_data` is
    #: provided).
    _subject: Unsettable[str | None]

    #: The first date/time in which the certificate is valid.
    #:
    #: If not provided during construction, this will be loaded from
    #: :py:attr`cert_data` when needed (and if :py:attr:`cert_data` is
    #: provided).
    _valid_from: Unsettable[datetime | None]

    #: The last date/time in which the certificate is valid.
    #:
    #: If not provided during construction, this will be loaded from
    #: :py:attr`cert_data` when needed (and if :py:attr:`cert_data` is
    #: provided).
    _valid_through: Unsettable[datetime | None]

    @classmethod
    def create_from_files(
        cls,
        *,
        hostname: str,
        port: int,
        cert_path: str,
        key_path: (str | None) = None,
        data_format: CertDataFormat = CertDataFormat.PEM,
        purpose: CertPurpose = CertPurpose.TRUST,
    ) -> Self:
        """Return an instance parsed from a PEM bundle file.

        Version Changed:
            8.0:
            * Added the ``purpose`` argument.
            * This may now raise :py:exc:`ValueError` if providing an invalid
              combination of arguments.

        Args:
            hostname (str):
                The primary hostname the certificate represents, used for
                lookup.

            port (int):
                The port that served the certificate.

            cert_path (str):
                The path to the local certificate file.

            key_path (str):
                The path to the local private key.

            data_format (CertDataFormat, optional):
                The format of the data stored in ``cert_path``.

            purpose (CertPurpose, optional):
                The purpose of the certificate.

                This sets whether the certificate should be used to trust
                a remote server or authenticate Review Board.

                By default, the purpose is to trust the server.

                Version Added:
                    8.0

        Raises:
            ValueError:
                One or more arguments were invalid.

                Version Added:
                    8.0

            reviewboard.certs.errors.CertificateNotFoundError:
                One or more of the certificate files was not found.

            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the CA bundle.

                Details are in the error message.
        """
        # Validate the arguments based on the purpose of the certificate.
        if purpose == CertPurpose.TRUST:
            if key_path:
                raise ValueError(_(
                    'key_path cannot be provided when passing '
                    'purpose=CertPurpose.TRUST.'
                ))
        elif purpose == CertPurpose.CLIENT:
            if not key_path:
                raise ValueError(_(
                    'key_path must be provided when passing '
                    'purpose=CertPurpose.CLIENT.'
                ))
        else:
            raise_invalid_type(
                purpose,
                _(
                    'Received an invalid value for purpose= ({purpose})'
                ).format(purpose=purpose),
            )

        # Validate the file paths.
        if not cert_path:
            raise ValueError(_(
                'A certificate path must be provided.'
            ))

        if not os.path.exists(cert_path):
            raise CertificateNotFoundError(
                _('The SSL/TLS certificate was not found.'))

        if key_path and not os.path.exists(key_path):
            raise CertificateNotFoundError(
                _('The SSL/TLS private key was not found.'))

        try:
            with open(cert_path, 'rb') as fp:
                cert_data = fp.read()
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Error reading SSL/TLS certificate file '
                         '"%s": %s',
                         error_id, cert_path, e)

            raise CertificateStorageError(
                _('Error reading SSL/TLS certificate file. Administrators can '
                  'find details in the Review Board server logs (error '
                  'ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

        if _CERT_PEM_RE.search(cert_data) is None:
            raise InvalidCertificateFormatError(data=cert_data,
                                                path=cert_path)

        if key_path is not None:
            try:
                with open(key_path, 'rb') as fp:
                    key_data = fp.read()
            except IOError as e:
                error_id = str(uuid4())
                logger.error('[%s] Error reading SSL/TLS private key file '
                             '"%s": %s',
                             error_id, cert_path, e)

                raise CertificateStorageError(
                    _('Error reading SSL/TLS private key file. Administrators '
                      'can find details in the Review Board server logs '
                      '(error ID %(error_id)s).')
                    % {
                        'error_id': error_id,
                    })

            if _PRIVATE_KEY_PEM_RE.search(key_data) is None:
                raise InvalidCertificateFormatError(data=key_data,
                                                    path=key_path)
        else:
            key_data = None

        return cls(hostname=hostname,
                   port=port,
                   cert_data=cert_data,
                   key_data=key_data,
                   purpose=purpose,
                   data_format=data_format)

    @classmethod
    def create_from_server(
        cls,
        *,
        hostname: str,
        port: int,
    ) -> Self | None:
        """Return a Certificate from a remote server.

        This is useful when inspecting a server or preparing an error
        response when certificate validation fails.

        Version Added:
            7.1

        Args:
            hostname (str):
                The hostname of the server.

            port (int):
                The port of the server.

        Returns:
            Certificate:
            The fetched certificate, or ``None`` if it could not be fetched.
        """
        try:
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

            with socket.create_connection((hostname, port),
                                          timeout=10) as sock:
                with context.wrap_socket(sock,
                                         server_hostname=hostname) as ssock:
                    der_cert = ssock.getpeercert(binary_form=True)

            if not der_cert:
                raise ValueError('Returned certificate was empty.')

            pem_cert = ssl.DER_cert_to_PEM_cert(der_cert).encode('ascii')

            return cls(
                hostname=hostname,
                port=port,
                cert_data=pem_cert,
            )
        except Exception as e:
            logger.warning('Failed to fetch server certificate from %s:%s: %s',
                           hostname, port, e)

            return None

    def __init__(
        self,
        *,
        hostname: str,
        port: int,
        cert_data: (bytes | None) = None,
        key_data: (bytes | None) = None,
        data_format: CertDataFormat = CertDataFormat.PEM,
        purpose: CertPurpose = CertPurpose.TRUST,
        fingerprints: Unsettable[CertificateFingerprints] = UNSET,
        issuer: Unsettable[str] = UNSET,
        subject: Unsettable[str] = UNSET,
        valid_from: Unsettable[datetime] = UNSET,
        valid_through: Unsettable[datetime] = UNSET,
    ) -> None:
        """Initialize the certificate.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname that would serve this certificate.

            port (int):
                The port on the host that would serve this certificate.

            cert_data (bytes):
                The loaded certificate data.

                This must be in the format defined by ``data_format``.

            key_data (bytes, optional):
                The loaded private key data, if available.

                This must be in the format defined by ``data_format``.

            data_format (CertDataFormat, optional):
                The format used for ``cert_data`` and ``key_data``.

                This currently only accepts PEM-encoded data.

            purpose (CertPurpose, optional):
                The purpose of the certificate.

                This sets whether the certificate should be used to trust
                a remote server or authenticate Review Board.

                By default, the purpose is to trust the server.

                Version Added:
                    8.0

            subject (str, optional):
                The subject (usually the hostname) of the certificate.

                If not provided, this will be loaded from ``cert_data`` when
                needed (and if ``cert_data`` is provided).

            issuer (str, optional):
                The issuer of the certificate.

                If not provided, this will be loaded from ``cert_data`` when
                needed (and if ``cert_data`` is provided).

            valid_from (datetime, optional):
                The first date/time in which the certificate is valid.

                This must have a timezone associated with it.

                If not provided, this will be loaded from ``cert_data`` when
                needed (and if ``cert_data`` is provided).

            valid_through (datetime, optional):
                The last date/time in which the certificate is valid.

                This must have a timezone associated with it.

                If not provided, this will be loaded from ``cert_data`` when
                needed (and if ``cert_data`` is provided).

            fingerprints (CertificateFingerprints, optional):
                Fingerprints to set for the certificate.

                If not provided, this will be loaded from ``cert_data`` when
                needed (and if ``cert_data`` is provided).

        Raises:
            ValueError:
                One or more arguments were invalid.
        """
        # Validate the arguments based on the purpose of the certificate.
        #
        # Note that this Certificate may represent an incomplete state, so
        # it may not have the cert/key data. However, if it does have cert
        # data, we can at least verify the key data against it.
        if purpose == CertPurpose.TRUST:
            if key_data:
                raise ValueError(_(
                    'key_data cannot be provided when passing '
                    'purpose=CertPurpose.TRUST.'
                ))
        elif purpose == CertPurpose.CLIENT:
            if cert_data and not key_data:
                raise ValueError(_(
                    'key_data must be provided when passing '
                    'purpose=CertPurpose.CLIENT.'
                ))
        else:
            raise_invalid_type(
                purpose,
                _(
                    'Received an invalid value for purpose= ({purpose})'
                ).format(purpose=purpose),
            )

        if valid_from is not UNSET and django_timezone.is_naive(valid_from):
            raise ValueError('valid_from must contain a timezone.')

        if (valid_through is not UNSET and
            django_timezone.is_naive(valid_through)):
            raise ValueError('valid_through must contain a timezone.')

        self.cert_data = cert_data
        self.data_format = data_format
        self.hostname = hostname
        self.key_data = key_data
        self.port = port
        self.purpose = purpose
        self._fingerprints = fingerprints
        self._issuer = issuer
        self._subject = subject
        self._valid_from = valid_from
        self._valid_through = valid_through

    @property
    def fingerprints(self) -> CertificateFingerprints | None:
        """Fingerprints for the certificate.

        Type:
            CertificateFingerprints
        """
        fingerprints = self._fingerprints

        if fingerprints is UNSET:
            x509_cert = self.x509_cert

            if x509_cert is None:
                fingerprints = None
            else:
                fingerprints = CertificateFingerprints.from_x509_cert(
                    x509_cert)

            self._fingerprints = fingerprints

        return fingerprints

    @cached_property
    def x509_cert(self) -> x509.Certificate | None:
        """A Cryptography X509 Certificate representing this certificate.

        This will be created from the loaded from the certificate data stored
        in :py:attr:`cert_data`. The created instance will be locally cached
        for future lookups.

        If certificate data is not available, this will be ``None``.

        Type:
            cryptography.x509.Certificate
        """
        if self.cert_data is None:
            return None

        assert self.data_format == CertDataFormat.PEM, (
            'Certificate must use PEM format.'
        )
        assert self.cert_data, (
            'Certificate data must be loaded.'
        )

        return x509.load_pem_x509_certificate(self.cert_data)

    @property
    def subject(self) -> str | None:
        """The subject of the certificate.

        Type:
            str
        """
        subject = self._subject

        if subject is UNSET:
            subject = self._get_x509_attr(str, 'subject')
            self._subject = subject

        return subject

    @cached_property
    def subject_alternative_names(self) -> Sequence[SANValue]:
        """The Subject Alternative Names of the certificate.

        This will include structured values for any DNS and IP addresses
        found in the Subject Alternative Names extension of the certificate,
        if present.

        Version Added:
            8.0

        Type:
            list of SANValue:
            A list of supported SAN values.
        """
        x509_cert = self.x509_cert

        if not x509_cert:
            return []

        try:
            san_ext = (
                x509_cert.extensions
                .get_extension_for_class(x509.SubjectAlternativeName)
            )

            return [
                item.value
                for item in san_ext.value
                if isinstance(item, (x509.DNSName, x509.IPAddress))
            ]
        except x509.ExtensionNotFound:
            # The extension wasn't found in the cert, so return an empty list.
            return []
        except Exception as e:
            logger.exception('Unexpected error retrieving SAN values for '
                             'x509 cert %r: %s',
                             x509_cert, e)

            return []

    @property
    def issuer(self) -> str | None:
        """The issuer of the certificate.

        Type:
            str
        """
        issuer = self._issuer

        if issuer is UNSET:
            issuer = self._get_x509_attr(str, 'issuer')
            self._issuer = issuer

        return issuer

    @property
    def valid_from(self) -> datetime | None:
        """The date/time in which the certificate is first valid.

        Type:
            datetime.datetime
        """
        valid_from = self._valid_from

        if valid_from is UNSET:
            valid_from = self._get_x509_attr(datetime, 'not_valid_before')
            self._valid_from = valid_from

        return valid_from

    @property
    def valid_through(self) -> datetime | None:
        """The last date/time in which the certificate is valid.

        Type:
            datetime.datetime
        """
        valid_through = self._valid_through

        if valid_through is UNSET:
            valid_through = self._get_x509_attr(datetime, 'not_valid_after')
            self._valid_through = valid_through

        return valid_through

    @property
    def is_valid(self) -> bool:
        """Whether this certificate is still considered valid.

        The certificate is valid if the current date/time is within its
        validity date range.

        Type:
            bool
        """
        valid_from = self.valid_from
        valid_through = self.valid_through

        return (valid_from is not None and
                valid_through is not None and
                valid_from <= django_timezone.now() <= valid_through)

    @property
    def is_wildcard(self) -> bool:
        """Whether this is a wildcard certificate.

        Wildcard certificates pertain to multiple domains (e.g.,
        ``*.example.com``, ``*a.example.com``, or ``b*.example.com``).

        Type:
            bool
        """
        return '*' in self.hostname

    def matches_host(
        self,
        hostname_or_ip: str,
    ) -> bool:
        """Return whether the certificate matches the provided hostname or IP.

        A certificate is a match if any of the following conditions are
        true:

        * The subject is a match for the hostname.
        * The first label of a cert hostname is a wildcard and matches the
          first label of the hostname, and the remaining labels are a match.
        * The value is an IPv4/IPv6 address and matches a SAN entry.

        Hostnames are all normalized for comparison.

        Partial wildcards (e.g., ``foo*.example.com``, ``*bar.example.com``, or
        ``foo*bar.example.com``) are not supported. Most Certificate
        Authorities no longer support these, and major browsers (including
        Chrome) consider them security risks.

        Version Added:
            8.0

        Args:
            hostname_or_ip (str):
                The hostname or IP address to check.

        Returns:
            bool:
            ``True`` if the host is matched by the certificate.
            ``False`` if it is not.
        """
        # Check if the provided hostname represents an IP address or a
        # hostname.
        #
        # If it's an IP address, we'll compare against any IPv4Address or
        # IPv6Address entries in the SAN field.
        try:
            check_ip = ip_address(hostname_or_ip)
        except ValueError:
            # This is not an IP address. Carry on with hostname checks.
            pass
        else:
            ip_cls = type(check_ip)

            # This is an IP address. Just compare with every equivalent
            # value in the SAN.
            return any(
                san_value == check_ip
                for san_value in self.subject_alternative_names
                if isinstance(san_value, ip_cls)
            )

        # This is a hostname. We'll check against the standard fields.
        check_hostname = normalize_cert_hostname(hostname_or_ip)

        # First, check against the Subject Common Name, if present.
        if subject := self.subject:
            subject = normalize_cert_hostname(subject)

            if get_cert_hostname_matches(cert_hostname=subject,
                                         check_hostname=check_hostname,
                                         normalize_hostnames=False):
                return True

        # Check all the Subject Alternative Name fields.
        for san_name in self.subject_alternative_names:
            if isinstance(san_name, str):
                # This is a string representing a hostname.
                san_name = normalize_cert_hostname(san_name)

                if get_cert_hostname_matches(cert_hostname=san_name,
                                             check_hostname=check_hostname,
                                             normalize_hostnames=False):
                    return True

        return False

    def to_json(self) -> SerializableDjangoJSONDictImmutable:
        """Serialize the certificate to data ready to be serialized to JSON.

        Returns:
            dict:
            The resulting JSON payload, containing:

            Keys:
                fingerprints (dict):
                    A dictionary of fingerprints for the certificate, or
                    ``None`` if not available.

                hostname (str):
                    The hostname serving the certificate.

                issuer (str):
                    The issuer of the certificate, or ``None`` if not
                    available.

                port (int):
                    The port on the host serving the certificate.

                purpose (str):
                    The purpose of the certificate.

                    Version Added:
                        8.0

                subject (str):
                    The subject of the certificate, or ``None`` if not
                    available.

                valid_from (str):
                    The first date/time in which the certificate is valid, or
                    ``None`` if not available.

                    This will be in :term:`ISO8601 format`.

                valid_through (str):
                    The last date/time in which the certificate is valid, or
                    ``None`` if not available.

                    This will be in :term:`ISO8601 format`.
        """
        return {
            'fingerprints': self.fingerprints,
            'hostname': self.hostname,
            'issuer': self.issuer,
            'port': self.port,
            'purpose': self.purpose.value,
            'subject': self.subject,
            'valid_from': self.valid_from,
            'valid_through': self.valid_through,
        }

    def write_cert_file(
        self,
        path: str,
    ) -> None:
        """Write the certificate data to a file.

        Args:
            path (str):
                The file path where the certificate data will be written.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error writing the file.
        """
        cert_data = self.cert_data

        assert cert_data, 'Cannot write empty certificate data to file.'

        try:
            with open(path, 'wb') as fp:
                fp.write(cert_data)
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Error writing SSL/TLS certificate file '
                         '"%s": %s',
                         error_id, path, e)

            raise CertificateStorageError(
                _('Error writing SSL/TLS certificate file. Administrators can '
                  'find details in the Review Board server logs (error '
                  'ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

    def write_key_file(
        self,
        path: str,
    ) -> None:
        """Write the private key data to a file.

        Args:
            path (str):
                The file path where the private key data will be written.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error writing the file.
        """
        key_data = self.key_data

        assert key_data, 'Cannot write empty certificate key data to file.'

        try:
            with open(path, 'wb') as fp:
                fp.write(key_data)
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Error writing SSL/TLS private key file '
                         '"%s": %s',
                         error_id, path, e)

            raise CertificateStorageError(
                _('Error writing SSL/TLS private key file. Administrators can '
                  'find details in the Review Board server logs (error '
                  'ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

    def _get_x509_attr(
        self,
        attr_type: type[_T],
        field_name: str,
    ) -> _T | None:
        """Return a normalized value for an X509.Certificate attribute.

        Any "Name" fields will be converted to a string.

        Any datetime fields will be made aware, using UTC.

        If the certificate could not be loaded, this will always return
        ``None``.

        Args:
            field_name (str):
                The name of the field on the certificate.

        Returns:
            object:
            The resulting value, or ``None``.
        """
        value: (_T | None) = None
        x509_cert = self.x509_cert

        if x509_cert is not None:
            x509_value = getattr(x509_cert, field_name)

            if isinstance(x509_value, x509.Name):
                x509_value = force_str(
                    x509_value
                    .get_attributes_for_oid(NameOID.COMMON_NAME)[0]
                    .value
                )
            elif (isinstance(x509_value, datetime) and
                  django_timezone.is_naive(x509_value)):
                x509_value = django_timezone.make_aware(x509_value,
                                                        timezone=timezone.utc)

            assert isinstance(x509_value, attr_type)

            value = x509_value

        return value

    def __repr__(self) -> str:
        """Return a string representation of the instance.

        Returns:
            str:
            The string representation.
        """
        return (
            f'<Certificate(hostname={self.hostname!r},'
            f' port={self.port!r},'
            f' purpose={self.purpose},'
            f' fingerprints={self.fingerprints!r})>'
        )


class CertificateBundle:
    """A bundle of root and intermediary certificates.

    This represents a "CA bundle," which specifies a root certificate and any
    necessary intermediary certificates used to validate other certificates,
    including those signed using an in-house certificate authority.

    Consumers should take care not to modify any certificate data after
    loading. While it's possible to change the data, doing so can lead to
    incorrect results, as some data is computed and then cached on the
    instance and cannot be updated later.

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The loaded data of the certificate bundle.
    bundle_data: bytes

    #: The format for the loaded certificate and private key data.
    data_format: CertDataFormat

    #: The name of this bundle.
    #:
    #: This is in :term:`slug` format.
    name: str

    @classmethod
    def create_from_file(
        cls,
        *,
        name: str,
        path: str,
    ) -> Self:
        """Return an instance parsed from a PEM bundle file.

        Args:
            name (str):
                The name of this bundle file.

                This must be in :term:`slug` format.

            path (str):
                The path to the file.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the CA bundle.

                Details are in the error message.
        """
        if not os.path.exists(path):
            raise CertificateNotFoundError(
                _('The SSL/TLS CA bundle was not found.'))

        try:
            with open(path, 'rb') as fp:
                data = fp.read()
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Error reading SSL/TLS CA bundle file '
                         '"%s": %s',
                         error_id, path, e)

            raise CertificateStorageError(
                _('Error loading SSL/TLS CA bundle file. Administrators can '
                  'find details in the Review Board server logs (error '
                  'ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

        if _CERT_PEM_RE.search(data) is None:
            raise InvalidCertificateFormatError(data=data,
                                                path=path)

        return cls(bundle_data=data,
                   data_format=CertDataFormat.PEM,
                   name=name)

    def __init__(
        self,
        *,
        bundle_data: bytes,
        data_format: CertDataFormat = CertDataFormat.PEM,
        name: str = 'certs',
    ) -> None:
        """Initialize the certificate bundle.

        Args:
            bundle_data (bytes):
                The loaded data of the certificate bundle.

            data_format (CertDataFormat, optional):
                The format used for ``contents``.

                This currently only accepts PEM-encoded data.

            name (str, optional):
                The name of the certificate bundle.
        """
        if name != slugify(name):
            raise ValueError(
                _('The certificate bundle name "%(name)s" must be in '
                  '"slug" format (using characters "a-z", "0-9", "-").')
                % {
                    'name': name,
                })

        self.bundle_data = bundle_data
        self.data_format = data_format
        self.name = name

    def write_bundle_file(
        self,
        path: str,
    ) -> None:
        """Write the certificate bundle data to a file.

        Args:
            path (str):
                The file path where the certificate bundle data will be
                written.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error writing the file.
        """
        try:
            with open(path, 'wb') as fp:
                fp.write(self.bundle_data)
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Error writing SSL/TLS CA bundle file '
                         '"%s": %s',
                         error_id, path, e)

            raise CertificateStorageError(
                _('Error writing SSL/TLS CA bundle file. Administrators can '
                  'find details in the Review Board server logs (error '
                  'ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })
